# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions for performing calculations on Gaussian states.
"""

from typing import Any, Optional, Sequence, Tuple, Union

from thewalrus.quantum import is_pure_cov

from mrmustard import math, settings
from mrmustard.math.tensor_wrappers.xptensor import XPMatrix, XPVector
from mrmustard.utils.typing import Matrix, Scalar, Vector

#  ~~~~~~
#  States
#  ~~~~~~


def vacuum_cov(num_modes: int) -> Matrix:
    r"""Returns the real covariance matrix of the vacuum state.

    Args:
        num_modes (int): number of modes

    Returns:
        Matrix: vacuum covariance matrix
    """
    return math.eye(num_modes * 2, dtype=math.float64) * settings.HBAR / 2


def vacuum_means(num_modes: int) -> Tuple[Matrix, Vector]:
    r"""Returns the real covariance matrix and real means vector of the vacuum state.

    Args:
        num_modes (int): number of modes

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    return displacement(
        math.zeros(num_modes, dtype="float64"),
        math.zeros(num_modes, dtype="float64"),
    )


def squeezed_vacuum_cov(r: Vector, phi: Vector) -> Matrix:
    r"""Returns the real covariance matrix and real means vector of a squeezed vacuum state.

    The dimension depends on the dimensions of ``r`` and ``phi``.

    Args:
        r (vector): squeezing magnitude
        phi (vector): squeezing angle

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    S = squeezing_symplectic(r, phi)
    return math.matmul(S, math.transpose(S)) * settings.HBAR / 2


def thermal_cov(nbar: Vector) -> Tuple[Matrix, Vector]:
    r"""Returns the real covariance matrix and real means vector of a thermal state.

    The dimension depends on the dimensions of ``nbar``.

    Args:
        nbar (vector): average number of photons per mode

    Returns:
        Matrix, Vector: thermal state covariance matrix or means vector
    """
    g = (2 * math.atleast_1d(nbar) + 1) * settings.HBAR / 2
    return math.diag(math.concat([g, g], axis=-1))


def two_mode_squeezed_vacuum_cov(r: Vector, phi: Vector) -> Matrix:
    r"""Returns the real covariance matrix and real means vector of a two-mode squeezed vacuum state.

    The dimension depends on the dimensions of ``r`` and ``phi``.

    Args:
        r (vector): squeezing magnitude
        phi (vector): squeezing angle

    Returns:
        Matrix: two-mode squeezed state covariance matrix
        Vector: two-mode squeezed state means vector
    """
    S = two_mode_squeezing_symplectic(r, phi)
    return math.matmul(S, math.transpose(S)) * settings.HBAR / 2


def gaussian_cov(symplectic: Matrix, eigenvalues: Vector = None) -> Matrix:
    r"""Returns the covariance matrix of a Gaussian state.

    Args:
        symplectic (Tensor): symplectic matrix of a channel
        eigenvalues (vector): symplectic eigenvalues

    Returns:
        Tensor: covariance matrix of the Gaussian state
    """
    if eigenvalues is None:
        return math.matmul(symplectic, math.transpose(symplectic))

    return math.matmul(
        math.matmul(symplectic, math.diag(math.concat([eigenvalues, eigenvalues], axis=0))),
        math.transpose(symplectic),
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~
#  Unitary transformations
# ~~~~~~~~~~~~~~~~~~~~~~~~


def rotation_symplectic(angle: Union[Scalar, Vector]) -> Matrix:
    r"""Symplectic matrix of a rotation gate.

    The dimension depends on the dimension of the angle.

    Args:
        angle (scalar or vector): rotation angles

    Returns:
        Tensor: symplectic matrix of a rotation gate
    """
    angle = math.atleast_1d(angle)
    num_modes = angle.shape[-1]
    x = math.cos(angle)
    y = math.sin(angle)
    return (
        math.diag(math.concat([x, x], axis=0))
        + math.diag(-y, k=num_modes)
        + math.diag(y, k=-num_modes)
    )


def squeezing_symplectic(r: Union[Scalar, Vector], phi: Union[Scalar, Vector]) -> Matrix:
    r"""Symplectic matrix of a squeezing gate.

    The dimension depends on the dimension of ``r`` and ``phi``.

    Args:
        r (scalar or vector): squeezing magnitude
        phi (scalar or vector): rotation parameter

    Returns:
        Tensor: symplectic matrix of a squeezing gate
    """
    r = math.atleast_1d(r, math.float64)
    phi = math.atleast_1d(phi, math.float64)
    if r.shape[-1] == 1:
        r = math.tile(r, phi.shape)
    if phi.shape[-1] == 1:
        phi = math.tile(phi, r.shape)
    num_modes = phi.shape[-1]
    cp = math.cos(phi)
    sp = math.sin(phi)
    ch = math.cosh(r)
    sh = math.sinh(r)
    cpsh = cp * sh
    spsh = sp * sh
    return (
        math.diag(math.concat([ch - cpsh, ch + cpsh], axis=0))
        + math.diag(-spsh, k=num_modes)
        + math.diag(-spsh, k=-num_modes)
    )


def displacement(x: Union[Scalar, Vector], y: Union[Scalar, Vector]) -> Vector:
    r"""Returns the displacement vector for a displacement by :math:`alpha = x + iy`.
    The dimension depends on the dimensions of ``x`` and ``y``.

    Args:
        x (scalar or vector): real part of displacement (in units of :math:`\sqrt{\hbar}`)
        y (scalar or vector): imaginary part of displacement (in units of :math:`\sqrt{\hbar}`)

    Returns:
        Vector: displacement vector of a displacement gate
    """
    x = math.atleast_1d(x, math.float64)
    y = math.atleast_1d(y, math.float64)
    if x.shape[-1] == 1:
        x = math.tile(x, y.shape)
    if y.shape[-1] == 1:
        y = math.tile(y, x.shape)
    return math.sqrt(2 * settings.HBAR, dtype=x.dtype) * math.concat([x, y], axis=0)


def beam_splitter_symplectic(theta: Scalar, phi: Scalar) -> Matrix:
    r"""Symplectic matrix of a Beam-splitter gate.

    The dimension is :math:`4\times 4`.

    Args:
        theta: transmissivity parameter
        phi: phase parameter

    Returns:
        Matrix: symplectic (orthogonal) matrix of a beam-splitter gate
    """
    ct = math.cos(theta)
    st = math.sin(theta)
    cp = math.cos(phi)
    sp = math.sin(phi)
    zero = math.zeros_like(theta)
    return math.astensor(
        [
            [ct, -cp * st, zero, -sp * st],
            [cp * st, ct, -sp * st, zero],
            [zero, sp * st, ct, -cp * st],
            [sp * st, zero, cp * st, ct],
        ]
    )


def mz_symplectic(phi_a: Scalar, phi_b: Scalar, internal: bool = False) -> Matrix:
    r"""Symplectic matrix of a Mach-Zehnder gate.

    It supports two conventions:

        * if ``internal=True``, both phases act inside the interferometer:
            ``phi_a`` on the upper arm, ``phi_b`` on the lower arm;
        * if `internal = False` (default), both phases act on the upper arm:
            ``phi_a`` before the first BS, ``phi_b`` after the first BS.

    Args:
        phi_a (float): first phase
        phi_b (float): second phase
        internal (bool): whether phases are in the internal arms (default is False)

    Returns:
        Matrix: symplectic (orthogonal) matrix of a Mach-Zehnder interferometer
    """
    ca = math.cos(phi_a)
    sa = math.sin(phi_a)
    cb = math.cos(phi_b)
    sb = math.sin(phi_b)
    cp = math.cos(phi_a + phi_b)
    sp = math.sin(phi_a + phi_b)

    if internal:
        return 0.5 * math.astensor(
            [
                [ca - cb, -sa - sb, sb - sa, -ca - cb],
                [-sa - sb, cb - ca, -ca - cb, sa - sb],
                [sa - sb, ca + cb, ca - cb, -sa - sb],
                [ca + cb, sb - sa, -sa - sb, cb - ca],
            ]
        )

    return 0.5 * math.astensor(
        [
            [cp - ca, -sb, sa - sp, -1 - cb],
            [-sa - sp, 1 - cb, -ca - cp, sb],
            [sp - sa, 1 + cb, cp - ca, -sb],
            [cp + ca, -sb, -sa - sp, 1 - cb],
        ]
    )


def two_mode_squeezing_symplectic(r: Scalar, phi: Scalar) -> Matrix:
    r"""Symplectic matrix of a two-mode squeezing gate.

    The dimension is :math:`4\times 4`.

    Args:
        r (float): squeezing magnitude
        phi (float): rotation parameter

    Returns:
        Matrix: symplectic matrix of a two-mode squeezing gate
    """
    cp = math.cast(math.cos(phi), math.float64)
    sp = math.cast(math.sin(phi), math.float64)
    ch = math.cast(math.cosh(r), math.float64)
    sh = math.cast(math.sinh(r), math.float64)
    zero = math.cast(math.zeros_like(math.asnumpy(r)), math.float64)
    return math.astensor(
        [
            [ch, cp * sh, zero, sp * sh],
            [cp * sh, ch, sp * sh, zero],
            [zero, sp * sh, ch, -cp * sh],
            [sp * sh, zero, -cp * sh, ch],
        ]
    )


def quadratic_phase(s: Scalar):
    r"""Quadratic phase single mode gate.

    .. math::

        P = \exp(i s q^2 / 2 \hbar)

    Reference: https://strawberryfields.ai/photonics/conventions/gates.html

    Args:
        s (float): interaction strength

    Returns:
        Tensor: the :math:`P(s)` matrix (in ``xxpp`` ordering)
    """

    return math.astensor(
        [
            [1, 0],
            [s, 1],
        ]
    )


def controlled_Z(g: Scalar):
    r"""Controlled PHASE gate of two-gaussian modes.

    .. math::

        C_Z = \exp(ig q_1 \otimes q_2 / \hbar).


    Reference: https://arxiv.org/pdf/2110.03247.pdf, Equation 8.
    https://arxiv.org/pdf/1110.3234.pdf, Equation 161.

    Args:
        g (float): interaction strength

    Returns:
        the C_Z(g) matrix (in xxpp ordering)
    """

    return math.astensor(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, g, 1, 0],
            [g, 0, 0, 1],
        ]
    )


def controlled_X(g: Scalar):
    r"""Controlled NOT gate of two-gaussian modes.

    .. math::

        C_X = \exp(ig q_1 \otimes p_2).

    Reference: https://arxiv.org/pdf/2110.03247.pdf, Equation 9.

    Args:
        g (float): interaction strength

    Returns:
        the C_X(g) matrix (in xxpp ordering)
    """

    return math.astensor(
        [
            [1, 0, 0, 0],
            [g, 1, 0, 0],
            [0, 0, 1, -g],
            [0, 0, 0, 1],
        ]
    )


# ~~~~~~~~~~~~~
# CPTP channels
# ~~~~~~~~~~~~~


def CPTP(
    cov: Matrix,
    means: Vector,
    X: Matrix,
    Y: Matrix,
    d: Vector,
    state_modes: Sequence[int],
    transf_modes: Sequence[int],
) -> Tuple[Matrix, Vector]:
    r"""Returns the cov matrix and means vector of a state after undergoing a CPTP channel.

    Computed as ``cov = X \cdot cov \cdot X^T + Y`` and ``d = X \cdot means + d``.

    If the channel is single-mode, ``modes`` can contain ``M`` modes to apply the channel to,
    otherwise it must contain as many modes as the number of modes in the channel.

    Args:
        cov (Matrix): covariance matrix
        means (Vector): means vector
        X (Matrix): the X matrix of the CPTP channel
        Y (Matrix): noise matrix of the CPTP channel
        d (Vector): displacement vector of the CPTP channel
        state_modes (Sequence[int]): modes the state is defined on
        transf_modes (Sequence[int]): modes on which the channel acts

    Returns:
        Tuple[Matrix, Vector]: the covariance matrix and the means vector of the state after the CPTP channel
    """
    if not set(transf_modes).issubset(state_modes):
        raise ValueError(
            f"The channel should act on a subset of the state modes ({transf_modes} is not a subset of {state_modes})"
        )
    # if single-mode channel, apply to all modes indicated in `modes`
    if X is not None and X.shape[-1] == 2:
        X = math.single_mode_to_multimode_mat(X, len(transf_modes))
    if Y is not None and Y.shape[-1] == 2:
        Y = math.single_mode_to_multimode_mat(Y, len(transf_modes))
    if d is not None and d.shape[-1] == 2:
        d = math.single_mode_to_multimode_vec(d, len(transf_modes))
    indices = [
        state_modes.index(i) for i in transf_modes
    ]  # TODO: do this when calling the method instead of here?
    cov = math.left_matmul_at_modes(X, cov, indices)
    cov = math.right_matmul_at_modes(cov, math.transpose(X), indices)
    cov = math.add_at_modes(cov, Y, indices)
    means = math.matvec_at_modes(X, means, indices)
    means = math.add_at_modes(means, d, indices)
    return cov, means


def loss_XYd(
    transmissivity: Union[Scalar, Vector], nbar: Union[Scalar, Vector]
) -> Tuple[Matrix, Matrix, None]:
    r"""Returns the ``X``, ``Y`` matrices and the ``d`` vector for the noisy loss (attenuator) channel.

    .. math::

        X = math.sqrt(transmissivity)
        Y = (1-transmissivity) * (2 * nbar + 1) * hbar / 2

    Reference: Alessio Serafini - Quantum Continuous Variables (5.77, p. 108)

    Args:
        transmissivity (float): value of the transmissivity, must be between 0 and 1
        nbar (float): photon number expectation value in the environment (0 for pure loss channel)

    Returns:
        Tuple[Matrix, Matrix, None]: the ``X``, ``Y`` matrices and the ``d`` vector for the noisy
        loss channel
    """
    if math.any(transmissivity < 0) or math.any(transmissivity > 1):
        raise ValueError("transmissivity must be between 0 and 1")
    x = math.sqrt(transmissivity)
    X = math.diag(math.concat([x, x], axis=0))
    y = (1 - transmissivity) * (2 * nbar + 1) * settings.HBAR / 2
    Y = math.diag(math.concat([y, y], axis=0))
    return X, Y, None


def amp_XYd(gain: Union[Scalar, Vector], nbar: Union[Scalar, Vector]) -> Matrix:
    r"""Returns the ``X``, ``Y`` matrices and the d vector for the noisy amplifier channel.

    .. math::

        X = math.sqrt(gain)
        Y = (gain-1) * (2 * nbar + 1) * hbar / 2

    Reference: Alessio Serafini - Quantum Continuous Variables (5.77, p. 111)

    The quantum limited amplifier channel is recovered for ``nbar = 0.0``.

    Args:
        gain (float): value of the gain > 1
        nbar (float): photon number expectation value in the environment (0 for quantum
            limited amplifier)

    Returns:
        Tuple[Matrix, Vector]: the ``X``, ``Y`` matrices and the ``d`` vector for the noisy
        amplifier channel.
    """
    if math.any(gain < 1):
        raise ValueError("Gain must be larger than 1")
    x = math.sqrt(gain)
    X = math.diag(math.concat([x, x], axis=0))
    y = (gain - 1) * (2 * nbar + 1) * settings.HBAR / 2
    Y = math.diag(math.concat([y, y], axis=0))
    return X, Y, None


def noise_Y(noise: Union[Scalar, Vector]) -> Matrix:
    r"""Returns the ``X``, ``Y`` matrices and the d vector for the additive noise channel ``(Y = noise * (\hbar / 2) * I)``

    Args:
        noise (float): number of photons in the thermal state

    Returns:
        Tuple[None, Matrix, None]: the ``X``, ``Y`` matrices and the ``d`` vector of the noise channel.
    """
    return math.diag(math.concat([noise, noise], axis=0)) * settings.HBAR / 2


def compose_channels_XYd(
    X1: Matrix, Y1: Matrix, d1: Vector, X2: Matrix, Y2: Matrix, d2: Vector
) -> Tuple[Matrix, Matrix, Vector]:
    r"""Returns the combined ``X``, ``Y``, and ``d`` for two CPTP channels.

    Args:
        X1 (Matrix): the ``X`` matrix of the first CPTP channel
        Y1 (Matrix): the ``Y`` matrix of the first CPTP channel
        d1 (Vector): the displacement vector of the first CPTP channel
        X2 (Matrix): the ``X`` matrix of the second CPTP channel
        Y2 (Matrix): the ``Y`` matrix of the second CPTP channel
        d2 (Vector): the displacement vector of the second CPTP channel

    Returns:
        Tuple[Matrix, Matrix, Vector]: the combined ``X``, ``Y``, and ``d`` matrices
    """
    if X1 is None:
        X = X2
    elif X2 is None:
        X = X1
    else:
        X = math.matmul(X2, X1)
    if Y1 is None:
        Y = Y2
    elif Y2 is None:
        Y = Y1
    else:
        Y = math.matmul(math.matmul(X2, Y1), X2) + Y2
    if d1 is None:
        d = d2
    elif d2 is None:
        d = d1
    else:
        d = math.matmul(X2, d1) + d2
    return X, Y, d


# ~~~~~~~~~~~~~~~
# non-TP channels
# ~~~~~~~~~~~~~~~


def general_dyne(
    cov: Matrix,
    means: Vector,
    proj_cov: Matrix,
    proj_means: Optional[Vector] = None,
    modes: Optional[Sequence[int]] = None,
) -> Tuple[Scalar, Matrix, Vector]:
    r"""Returns the results of a general-dyne measurement. If ``proj_means`` are not provided
    (as ``None``), they are sampled from the probability distribution.

    Args:
        cov (Matrix): covariance matrix of the state being measured [units of `2\hbar`]
        means (Vector): means vector of the state being measured [units of `\sqrt(\hbar)`]
        proj_cov (Matrix): covariance matrix of the state being projected onto [units of `2\hbar`]
        proj_means (Optional Vector): means vector of the state being projected onto
            (i.e. the measurement outcome) [units of `\sqrt(\hbar)`]. If not provided, the means vector
            is sampled from the generaldyne probability distribution.
        modes (Optional Sequence[int]): modes being measured (modes are indexed from 0 to num_modes-1),
            if modes are not provided then the first modes (according to the size of ``cov``) are measured.

    Returns:
        Tuple[Scalar, Scalar, Matrix, Vector]:
            outcome (sampled means vector of the measured subsystem) [units of `\sqrt(\hbar)`],
            oucome probability [units of `\hbar**N`],
            post-measurement covariace [units of `2\hbar`]
            post-measurement means vector [units of `\sqrt{\hbar}`].
    """
    N, M = cov.shape[-1] // 2, proj_cov.shape[-1] // 2
    # Bmodes are the modes being measured and Amodes are the leftover modes
    Bmodes = modes or list(range(M))
    Amodes = list(set(list(range(N))) - set(Bmodes))

    A, B, AB = partition_cov(cov, Amodes)
    a, b = partition_means(means, Amodes)
    reduced_cov = B + proj_cov

    # covariances are divided by 2 to match tensorflow and MrMustard conventions
    # (MrMustard uses Serafini convention where `sigma_MM = 2 sigma_TF`)
    pdf = math.MultivariateNormalTriL(loc=b, scale_tril=math.cholesky(reduced_cov / 2))
    outcome = (
        pdf.sample(dtype=cov.dtype) if proj_means is None else math.cast(proj_means, cov.dtype)
    )
    prob = pdf.prob(outcome)

    # calculate conditional output state of unmeasured modes
    num_remaining_modes = N - M
    if num_remaining_modes == 0:
        return outcome, prob, None, None

    AB_inv = math.matmul(AB, math.inv(reduced_cov))
    new_cov = A - math.matmul(AB_inv, math.transpose(AB))
    new_means = a + math.matvec(AB_inv, outcome - b)

    return outcome, prob, new_cov, new_means


# ~~~~~~~~~
# utilities
# ~~~~~~~~~
def number_means(cov: Matrix, means: Vector) -> Vector:
    r"""Returns the photon number means vector given a Wigner covariance matrix and a means vector.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector

    Returns:
        Vector: the photon number means vector
    """
    N = means.shape[-1] // 2
    return (
        means[:N] ** 2
        + means[N:] ** 2
        + math.diag_part(cov[:N, :N])
        + math.diag_part(cov[N:, N:])
        - settings.HBAR
    ) / (2 * settings.HBAR)


def number_cov(cov: Matrix, means: Vector) -> Matrix:
    r"""Returns the photon number covariance matrix given a Wigner covariance matrix and a means vector.

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector

    Returns:
        Matrix: the photon number covariance matrix
    """
    N = means.shape[-1] // 2
    mCm = cov * means[:, None] * means[None, :]
    dd = math.diag(math.diag_part(mCm[:N, :N] + mCm[N:, N:] + mCm[:N, N:] + mCm[N:, :N])) / (
        2 * settings.HBAR**2  # TODO: sum(diag_part) is better than diag_part(sum)
    )
    CC = (cov**2 + mCm) / (2 * settings.HBAR**2)
    return (
        CC[:N, :N] + CC[N:, N:] + CC[:N, N:] + CC[N:, :N] + dd - 0.25 * math.eye(N, dtype=CC.dtype)
    )


def is_mixed_cov(cov: Matrix) -> bool:  # TODO: deprecate
    r"""Returns ``True`` if the covariance matrix is mixed, ``False`` otherwise."""
    return not is_pure_cov(math.asnumpy(cov))


def trace(cov: Matrix, means: Vector, Bmodes: Sequence[int]) -> Tuple[Matrix, Vector]:
    r"""Returns the covariances and means after discarding the specified modes.

    Args:
        cov (Matrix): covariance matrix
        means (Vector): means vector
        Bmodes (Sequence[int]): modes to discard

    Returns:
        Tuple[Matrix, Vector]: the covariance matrix and the means vector after discarding the specified modes
    """
    N = len(cov) // 2
    Aindices = math.astensor(
        [i for i in range(N) if i not in Bmodes] + [i + N for i in range(N) if i not in Bmodes],
        dtype=math.int32,
    )
    A_cov_block = math.gather(math.gather(cov, Aindices, axis=0), Aindices, axis=1)
    A_means_vec = math.gather(means, Aindices)
    return A_cov_block, A_means_vec


def partition_cov(cov: Matrix, Amodes: Sequence[int]) -> Tuple[Matrix, Matrix, Matrix]:
    r"""Partitions the covariance matrix into the ``A`` and ``B`` subsystems and the AB coherence block.

    Args:
        cov (Matrix): the covariance matrix
        Amodes (Sequence[int]): the modes of system ``A``

    Returns:
        Tuple[Matrix, Matrix, Matrix]: the cov of ``A``, the cov of ``B`` and the AB block
    """
    N = cov.shape[-1] // 2
    Bindices = math.cast(
        [i for i in range(N) if i not in Amodes] + [i + N for i in range(N) if i not in Amodes],
        "int32",
    )
    Aindices = math.cast(Amodes + [i + N for i in Amodes], "int32")
    A_block = math.gather(math.gather(cov, Aindices, axis=1), Aindices, axis=0)
    B_block = math.gather(math.gather(cov, Bindices, axis=1), Bindices, axis=0)
    AB_block = math.gather(math.gather(cov, Bindices, axis=1), Aindices, axis=0)
    return A_block, B_block, AB_block


def partition_means(means: Vector, Amodes: Sequence[int]) -> Tuple[Vector, Vector]:
    r"""Partitions the means vector into the ``A`` and ``B`` subsystems.

    Args:
        means (Vector): the means vector
        Amodes (Sequence[int]): the modes of system ``A``

    Returns:
        Tuple[Vector, Vector]: the means of ``A`` and the means of ``B``
    """
    N = len(means) // 2
    Bindices = math.cast(
        [i for i in range(N) if i not in Amodes] + [i + N for i in range(N) if i not in Amodes],
        "int32",
    )
    Aindices = math.cast(Amodes + [i + N for i in Amodes], "int32")
    return math.gather(means, Aindices), math.gather(means, Bindices)


def purity(cov: Matrix) -> Scalar:
    r"""Returns the purity of the state with the given covariance matrix.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the purity
    """
    return 1 / math.sqrt(math.det((2 / settings.HBAR) * cov))


def symplectic_eigenvals(cov: Matrix) -> Any:
    r"""Returns the sympletic eigenspectrum of a covariance matrix.

    For a pure state, we expect the sympletic eigenvalues to be 1.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        List[float]: the sympletic eigenvalues
    """
    J = math.J(cov.shape[-1] // 2)  # create a sympletic form
    M = math.matmul(1j * J, cov * (2 / settings.HBAR))
    vals = math.eigvals(M)  # compute the eigenspectrum
    return math.abs(vals[::2])  # return the even eigenvalues  # TODO: sort?


def von_neumann_entropy(cov: Matrix) -> float:
    r"""Returns the Von Neumann entropy.

    For a pure state, we expect the Von Neumann entropy to be 0.

    Reference: (https://arxiv.org/pdf/1110.3234.pdf), Equations 46-47.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the Von Neumann entropy
    """

    def g(x):
        return math.xlogy((x + 1) / 2, (x + 1) / 2) - math.xlogy((x - 1) / 2, (x - 1) / 2 + 1e-9)

    symp_vals = symplectic_eigenvals(cov)
    entropy = math.sum(g(symp_vals))
    return entropy


def fidelity(mu1: Vector, cov1: Matrix, mu2: Vector, cov2: Matrix) -> float:
    r"""Returns the fidelity of two gaussian states.

    Reference: `arXiv:2102.05748 <https://arxiv.org/pdf/2102.05748.pdf>`_, equations 95-99.
    Note that we compute the square of equation 98.

    Args:
        mu1 (Vector): the means vector of state 1
        mu2 (Vector): the means vector of state 2
        cov1 (Matrix): the covariance matrix of state 1
        cov1 (Matrix): the covariance matrix of state 2

    Returns:
        float: the fidelity
    """

    cov1 = math.cast(cov1 / settings.HBAR, "complex128")  # convert to units where hbar = 1
    cov2 = math.cast(cov2 / settings.HBAR, "complex128")  # convert to units where hbar = 1

    mu1 = math.cast(mu1, "complex128")
    mu2 = math.cast(mu2, "complex128")
    deltar = (mu2 - mu1) / math.sqrt(
        settings.HBAR, dtype=mu1.dtype
    )  # convert to units where hbar = 1
    J = math.J(cov1.shape[0] // 2)
    I = math.eye(cov1.shape[0])
    J = math.cast(J, "complex128")
    I = math.cast(I, "complex128")

    cov12_inv = math.inv(cov1 + cov2)

    V = math.transpose(J) @ cov12_inv @ ((1 / 4) * J + cov2 @ J @ cov1)

    W = -2 * (V @ (1j * J))
    W_inv = math.inv(W)
    matsqrtm = math.sqrtm(
        I - W_inv @ W_inv
    )  # this also handles the case where the input matrix is close to zero
    f0_top = math.det((matsqrtm + I) @ (W @ (1j * J)))
    f0_bot = math.det(cov1 + cov2)

    f0 = math.sqrt(f0_top / f0_bot)  # square of equation 98

    dot = math.sum(
        math.transpose(deltar) * math.matvec(cov12_inv, deltar)
    )  # computing (mu2-mu1)/sqrt(hbar).T @ cov12_inv @ (mu2-mu1)/sqrt(hbar)

    _fidelity = f0 * math.exp((-1 / 2) * dot)  # square of equation 95

    return math.real(_fidelity)


def physical_partial_transpose(cov: Matrix, modes: Sequence[int]) -> Matrix:
    r"""Returns the covariance matrix that corresponds to applying the partial
    transposition on the density matrix of a given set of modes.

    Reference: `https://arxiv.org/abs/quant-ph/9909044 <https://arxiv.org/abs/quant-ph/9909044>`_, Equation 1, 5.

    Args:
        cov (Matrix): the covariance matrix
        modes (Sequence[int]): the modes of system on which transposition is applied

    Returns:
        Matrix: the covariance matrix corresponding to the partially transposed state
    """
    m, _ = cov.shape
    num_modes = m // 2
    mat = [1.0] * m
    for i in modes:
        mat[i + num_modes] = -1.0
    mat = math.astensor(mat, dtype="float64")
    return cov * mat[:, None] * mat[None, :]


def log_negativity(cov: Matrix) -> float:
    r"""Returns the log_negativity of a Gaussian state.

    Reference: `https://arxiv.org/abs/quant-ph/0102117 <https://arxiv.org/abs/quant-ph/0102117>`_ , Equation 57, 61.

    Args:
        cov (Matrix): the covariance matrix

    Returns:
        float: the log-negativity
    """
    vals = symplectic_eigenvals(cov) / (settings.HBAR / 2)
    vals_filtered = math.boolean_mask(
        vals, vals < 1.0
    )  # Get rid of terms that would lead to zero contribution.
    if len(vals_filtered) > 0:
        return -math.sum(
            math.log(vals_filtered) / math.cast(math.log(2.0), dtype=vals_filtered.dtype)
        )

    return 0


def join_covs(covs: Sequence[Matrix]) -> Tuple[Matrix, Vector]:
    r"""Joins the given covariance matrices into a single covariance matrix.

    Args:
        covs (Sequence[Matrix]): the covariance matrices

    Returns:
        Matrix: the joined covariance matrix
    """
    modes = list(range(len(covs[0]) // 2))
    cov = XPMatrix.from_xxpp(covs[0], modes=(modes, modes), like_1=True)
    for _, c in enumerate(covs[1:]):
        modes = list(range(cov.num_modes, cov.num_modes + c.shape[-1] // 2))
        cov = cov @ XPMatrix.from_xxpp(c, modes=(modes, modes), like_1=True)
    return cov.to_xxpp()


def join_means(means: Sequence[Vector]) -> Vector:
    r"""Joins the given means vectors into a single means vector.

    Args:
        means (Sequence[Vector]): the means vectors

    Returns:
        Vector: the joined means vector
    """
    mean = XPVector.from_xxpp(means[0], modes=list(range(len(means[0]) // 2)))
    for _, m in enumerate(means[1:]):
        mean = mean + XPVector.from_xxpp(
            m, modes=list(range(mean.num_modes, mean.num_modes + len(m) // 2))
        )
    return mean.to_xxpp()


def symplectic_inverse(S: Matrix) -> Matrix:
    r"""Returns the inverse of a symplectic matrix.

    Args:
        S (Matrix): the symplectic matrix

    Returns:
        Matrix: the inverse of the symplectic matrix
    """
    S = math.reshape(S, (S.shape[0] // 2, 2, S.shape[1] // 2, 2))
    S = math.transpose(S, (1, 3, 0, 2))
    return math.block(
        [
            [math.transpose(S[1, 1]), -math.transpose(S[0, 1])],
            [-math.transpose(S[1, 0]), math.transpose(S[0, 0])],
        ]
    )


def XYd_dual(X: Matrix, Y: Matrix, d: Vector):
    r"""Returns the dual channel ``(X, Y, d)``.

    Args:
        X (Matrix): the ``X`` matrix
        Y (Matrix): the ``Y`` noise matrix
        d (Vector): the displacement vector

    Returns:
        Tuple[Matrix, Matrix, Vector]: ``(X_dual, Y_dual, d_dual)``
    """
    X_dual = math.inv(X) if X is not None else None
    Y_dual = Y
    d_dual = d
    if Y is not None:
        Y_dual = (
            math.matmul(X_dual, math.matmul(Y, math.transpose(X_dual))) if X_dual is not None else Y
        )
    if d is not None:
        d_dual = math.matvec(X_dual, d) if X_dual is not None else d
    return X_dual, Y_dual, d_dual



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sorting functions"""

from collections import defaultdict
from typing import Generator

import numpy as np


def argsort_gen(generators: list[Generator[float, None, None]]) -> list[int]:
    r"""
    Sorts a list of generator objects based on their yielded values.

    This function takes a list of generator objects, each yielding a sequence of numbers.
    It sorts the generators based on their first yielded values. If multiple generators
    yield the same first value, the function recursively sorts them based on their next
    yielded values. The result is a list of indices that would sort the original list of
    generators.

    Parameters:
    gen_list (list): A list of generator objects, each yielding a sequence of numbers.

    Returns:
    list: A list of indices that would sort the original list of generators.
    """
    vals = []
    for gen in generators:
        try:
            vals.append(next(gen))
        except StopIteration:
            vals.append(np.inf)

    if np.allclose(vals, np.inf):
        return [i for i in range(len(generators))]

    # dict with values and indices where they occur
    vals_dict = defaultdict(list)
    for i, val in enumerate(vals):
        vals_dict[val].append(i)

    # sort dict by keys (vals)
    vals_dict = {key: vals_dict[key] for key in sorted(vals_dict)}

    # if there are multiple values in the same pool, sort them with argsort_gen
    for val, index_pool in vals_dict.items():
        if len(index_pool) > 1:
            sub_gen_list = [generators[i] for i in index_pool]
            sub_order = argsort_gen(sub_gen_list)
            pool_sorted = [index_pool[i] for i in sub_order]
            vals_dict[val] = pool_sorted

    return [i for pool in vals_dict.values() for i in pool]



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name

"""
This module contains the implementation of a tensor wrapper class.
"""
import string
from numbers import Number
from typing import List, Optional, Union

from mrmustard.math.backend_manager import BackendManager

math = BackendManager()


class MMTensor:
    r"""A Mr Mustard tensor (a wrapper around an array that implements the numpy array API)."""

    def __init__(self, array, axis_labels=None):
        # If the input array is an MMTensor,
        # use its tensor and axis labels (or the provided ones if specified)
        if isinstance(array, MMTensor):
            self.tensor = array.tensor
            self.axis_labels = axis_labels or array.axis_labels
        else:
            self.tensor = array
            self.axis_labels = axis_labels

        # If axis labels are not provided, generate default labels
        if self.axis_labels is None:
            self.axis_labels = [str(n) for n in range(len(self.tensor.shape))]

        # Validate the number of axis labels
        if len(self.axis_labels) != len(self.tensor.shape):
            raise ValueError("The number of axis labels must be equal to the number of axes.")

    def __array__(self):
        r"""Implement the NumPy array interface."""
        return self.tensor

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        r"""
        Implement the NumPy ufunc interface.
        """
        if method == "__call__":
            inputs = [i.tensor if isinstance(i, MMTensor) else i for i in inputs]
            return MMTensor(ufunc(*inputs, **kwargs), self.axis_labels)
        else:
            return NotImplemented(f"Cannot call {method} on {ufunc}.")

    def __mul__(self, other):
        r"""implement the * operator"""
        if isinstance(other, Number):
            return MMTensor(self.tensor * other, self.axis_labels)
        if isinstance(other, MMTensor):
            self.check_axis_labels_match(other)
            return MMTensor(self.tensor * other.tensor, self.axis_labels)
        return NotImplemented(f"Cannot multiply {type(self)} and {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        r"""implement the / operator"""
        if isinstance(other, MMTensor):
            self.check_axis_labels_match(other)
            return MMTensor(self.tensor / other.tensor, self.axis_labels)
        try:
            return MMTensor(self.tensor / other, self.axis_labels)
        except TypeError:
            return NotImplemented(f"Cannot divide {type(self)} by {type(other)}")

    def __add__(self, other):
        r"""implement the + operator"""
        if isinstance(other, MMTensor):
            self.check_axis_labels_match(other)
            return MMTensor(self.tensor + other.tensor, self.axis_labels)
        try:
            return MMTensor(self.tensor + other, self.axis_labels)
        except TypeError:
            return NotImplemented(f"Cannot add {type(self)} and {type(other)}")

    def __sub__(self, other):
        return self.__add__(-other)

    def __neg__(self):
        return MMTensor(-self.tensor, self.axis_labels)

    def check_axis_labels_match(self, other):
        r"""
        Check that the axis labels of *this* tensor match those of another tensor.
        """
        if self.axis_labels != other.axis_labels:
            raise ValueError(
                f"Axis labels must match (got {self.axis_labels} and {other.axis_labels})"
            )

    def __radd__(self, other):
        return self.__add__(other)

    def __matmul__(self, other):
        r"""Overload the @ operator to perform tensor contractions."""
        # if not isinstance(other, MMTensor):
        #     raise TypeError(f"Cannot contract with object of type {type(other)}")

        # Find common axis labels
        common_labels = set(self.axis_labels) & set(other.axis_labels)

        if not common_labels:
            raise ValueError("No common axis labels found")

        # Determine the indices to contract along
        left_indices = [self.axis_labels.index(label) for label in common_labels]
        right_indices = [other.axis_labels.index(label) for label in common_labels]

        # Create a list of the new axis labels
        new_axis_labels = [label for label in self.axis_labels if label not in common_labels] + [
            label for label in other.axis_labels if label not in common_labels
        ]

        return MMTensor(
            math.tensordot(self.tensor, other.tensor, axes=(left_indices, right_indices)),
            new_axis_labels,
        )

    def contract(self, relabeling: Optional[List[str]] = None):
        r"""
        Contract *this* tensor along the specified indices using einsum.

        Args:
            relabeling (list[str]): An optional list of new axis labels.
            The tensor is contracted along all groups of axes with matching labels.
        """
        # check that labels are valid
        if relabeling is None:
            relabeling = self.axis_labels
        elif len(relabeling) != len(self.axis_labels):
            raise ValueError("The number of labels must be equal to the number of axes.")

        self.axis_labels = relabeling

        # Find all unique labels but keep the order
        unique_labels = []
        for label in relabeling:
            if label not in unique_labels:
                unique_labels.append(label)
        repeated = [label for label in unique_labels if self.axis_labels.count(label) > 1]

        # Turn labels into consecutive ascii lower-case letters,
        # with same letters corresponding to the same label
        label_map = {label: string.ascii_lowercase[i] for i, label in enumerate(unique_labels)}
        labels = [label_map[label] for label in self.axis_labels]

        # create einsum string from labels
        einsum_str = "".join(labels)

        # Contract the tensor and assign new axis labels (unique except for the contracted ones)
        return MMTensor(
            math.einsum(einsum_str, self.tensor),
            [label for label in unique_labels if label not in repeated],
        )

    def transpose(self, perm: Union[List[int], List[str]]):
        """Transpose the tensor using a list of axis labels or indices."""
        if set(perm) == set(self.axis_labels):
            perm = [self.axis_labels.index(label) for label in perm]
        return MMTensor(math.transpose(self.tensor, perm), [self.axis_labels[i] for i in perm])

    def reshape(self, shape, axis_labels=None):
        """Reshape the tensor. Allows to change the axis labels."""
        return MMTensor(math.reshape(self.tensor, shape), axis_labels or self.axis_labels)

    def __getitem__(self, indices):
        """Implement indexing into the tensor."""
        indices = indices if isinstance(indices, tuple) else (indices,)
        axis_labels = self.axis_labels.copy()
        offset = 0
        for i, ind in enumerate(indices):
            if isinstance(ind, int):
                axis_labels.pop(i + offset)
                offset -= 1
            elif ind is Ellipsis and i == 0:
                offset = len(self.tensor.shape) - len(indices)
            elif ind is Ellipsis and i > 0:
                break

        return MMTensor(self.tensor[indices], axis_labels)

    def __repr__(self):
        return f"MMTensor({self.tensor}, {self.axis_labels})"

    def __getattribute__(self, name):
        r"""
        Overrides built-in getattribute method for fallback attribute lookup.
        Tries to get attribute from self, then from self.tensor

        Args:
            self (object): instance
            name (str): attribute name

        Returns:
            attribute value or raises AttributeError"""
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.tensor, name)



"""
This module calculates the derivatives of the diagonal of the Fock representation (i.e. the PNR detection probabilities of all modes)
by applying the derivated recursion relation in a selective manner.
"""

import numpy as np
import numba
from numba import njit, int64
from numba.cpython.unsafe.tuple import tuple_setitem
from mrmustard.math.lattice.strategies.compactFock.helperFunctions import (
    SQRT,
    repeat_twice,
    construct_dict_params,
)


@njit
def calc_dA_dB(i, G_in_dA, G_in_dB, G_in, A, B, K_l, K_i, M, pivot_val, pivot_val_dA, pivot_val_dB):
    """
    Calculate the derivatives of one Fock amplitude w.r.t A and B.
    Args:
        i (int): the element of the multidim index that is increased
        G_in, G_in_dA, G_in_dB (array, array, array): all Fock amplitudes from the 'read' group in the recurrence relation and their derivatives w.r.t. A and B
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        K_l, K_i (vector, vector): SQRT[pivot], SQRT[pivot + 1]
        M (int): number of modes
        pivot_val, pivot_val_dA, pivot_val_dB (array, array, array): Fock amplitude at the position of the pivot and its derivatives w.r.t. A and B
    """
    dA = pivot_val_dA * B[i]
    dB = pivot_val_dB * B[i]
    dB[i] += pivot_val
    for l in range(2 * M):
        dA += K_l[l] * A[i, l] * G_in_dA[l]
        dB += K_l[l] * A[i, l] * G_in_dB[l]
        dA[i, l] += G_in[l]
    return dA / K_i[i], dB / K_i[i]


@njit
def use_offDiag_pivot_grad(
    A,
    B,
    M,
    cutoffs,
    params,
    d,
    arr0,
    arr2,
    arr1010,
    arr1001,
    arr1,
    arr0_dA,
    arr2_dA,
    arr1010_dA,
    arr1001_dA,
    arr1_dA,
    arr0_dB,
    arr2_dB,
    arr1010_dB,
    arr1001_dB,
    arr1_dB,
):
    """
    Apply recurrence relation for pivot of type [a+1,a,b,b,c,c,...] / [a,a,b+1,b,c,c,...] / [a,a,b,b,c+1,c,...]
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        d (int): mode index in which the considered Fock amplitude is off diagonal
            e.g. [a,a,b+1,b,c,c,...] --> b is off diagonal --> d=1
        arr0, arr2, arr1010, arr1001, arr1 (array, array, array, array, array): submatrices of the fock representation
        arr..._dA, arr..._dB (array, array): derivatives of submatrices w.r.t A and B
    Returns:
        (array, array, array, array, array): updated versions of arr0, arr2, arr1010, arr1001, arr1
    """
    pivot = repeat_twice(params)
    pivot[2 * d] += 1
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros(2 * M, dtype=np.complex128)
    G_in_dA = np.zeros((2 * M,) + A.shape, dtype=np.complex128)
    G_in_dB = np.zeros((2 * M,) + B.shape, dtype=np.complex128)

    ########## READ ##########
    pivot_val = arr1[2 * d][params]
    pivot_val_dA = arr1_dA[2 * d][params]
    pivot_val_dB = arr1_dB[2 * d][params]

    # Array0
    G_in[2 * d] = arr0[params]
    G_in_dA[2 * d] = arr0_dA[params]
    G_in_dB[2 * d] = arr0_dB[params]

    # read from Array2
    if params[d] > 0:
        G_in[2 * d + 1] = arr2[d][tuple_setitem(params, d, params[d] - 1)]
        G_in_dA[2 * d + 1] = arr2_dA[d][tuple_setitem(params, d, params[d] - 1)]
        G_in_dB[2 * d + 1] = arr2_dB[d][tuple_setitem(params, d, params[d] - 1)]

    # read from Array11
    for i in range(d + 1, M):  # i>d
        if params[i] > 0:
            params_adapted = tuple_setitem(params, i, params[i] - 1)
            G_in[2 * i] = arr1001[d][i - d - 1][params_adapted]
            G_in_dA[2 * i] = arr1001_dA[d][i - d - 1][params_adapted]
            G_in_dB[2 * i] = arr1001_dB[d][i - d - 1][params_adapted]
            G_in[2 * i + 1] = arr1010[d][i - d - 1][params_adapted]
            G_in_dA[2 * i + 1] = arr1010_dA[d][i - d - 1][params_adapted]
            G_in_dB[2 * i + 1] = arr1010_dB[d][i - d - 1][params_adapted]

    ########## WRITE ##########
    G_in = np.multiply(K_l, G_in)

    # Array0
    params_adapted = tuple_setitem(params, d, params[d] + 1)
    arr0_dA[params_adapted], arr0_dB[params_adapted] = calc_dA_dB(
        2 * d + 1, G_in_dA, G_in_dB, G_in, A, B, K_l, K_i, M, pivot_val, pivot_val_dA, pivot_val_dB
    )

    # Array2
    if params[d] + 2 < cutoffs[d]:
        arr2_dA[d][params], arr2_dB[d][params] = calc_dA_dB(
            2 * d, G_in_dA, G_in_dB, G_in, A, B, K_l, K_i, M, pivot_val, pivot_val_dA, pivot_val_dB
        )

    # Array11
    for i in range(d + 1, M):
        if params[i] + 1 < cutoffs[i]:
            arr1010_dA[d][i - d - 1][params], arr1010_dB[d][i - d - 1][params] = calc_dA_dB(
                2 * i,
                G_in_dA,
                G_in_dB,
                G_in,
                A,
                B,
                K_l,
                K_i,
                M,
                pivot_val,
                pivot_val_dA,
                pivot_val_dB,
            )
            arr1001_dA[d][i - d - 1][params], arr1001_dB[d][i - d - 1][params] = calc_dA_dB(
                2 * i + 1,
                G_in_dA,
                G_in_dB,
                G_in,
                A,
                B,
                K_l,
                K_i,
                M,
                pivot_val,
                pivot_val_dA,
                pivot_val_dB,
            )

    return arr0_dA, arr2_dA, arr1010_dA, arr1001_dA, arr0_dB, arr2_dB, arr1010_dB, arr1001_dB


@njit
def use_diag_pivot_grad(A, B, M, cutoffs, params, arr0, arr1, arr0_dA, arr1_dA, arr0_dB, arr1_dB):
    """
    Apply recurrence relation for pivot of type [a,a,b,b,c,c...]
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        arr0, arr1 (array, array): submatrices of the fock representation
        arr..._dA, arr..._dB (array, array): derivatives of submatrices w.r.t A and B
    Returns:
        (array, array): updated versions of arr0, arr1
    """
    pivot = repeat_twice(params)
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros(2 * M, dtype=np.complex128)
    G_in_dA = np.zeros((2 * M,) + A.shape, dtype=np.complex128)
    G_in_dB = np.zeros((2 * M,) + B.shape, dtype=np.complex128)

    ########## READ ##########
    pivot_val = arr0[params]
    pivot_val_dA = arr0_dA[params]
    pivot_val_dB = arr0_dB[params]

    # Array1
    for i in range(2 * M):
        if params[i // 2] > 0:
            i_staggered = i + 1 - 2 * (i % 2)  # [i+1-2*(i%2) for i in range(6)] == [1,0,3,2,5,4]
            params_adapted = tuple_setitem(params, i // 2, params[i // 2] - 1)
            G_in[i] = arr1[i_staggered][params_adapted]
            G_in_dA[i] = arr1_dA[i_staggered][params_adapted]
            G_in_dB[i] = arr1_dB[i_staggered][params_adapted]

    ########## WRITE ##########
    G_in = np.multiply(K_l, G_in)

    # Array1
    for i in range(2 * M):
        if params[i // 2] + 1 < cutoffs[i // 2]:
            # this if statement prevents a few elements from being written that will never be read
            if i != 1 or params[0] + 2 < cutoffs[0]:
                arr1_dA[i][params], arr1_dB[i][params] = calc_dA_dB(
                    i,
                    G_in_dA,
                    G_in_dB,
                    G_in,
                    A,
                    B,
                    K_l,
                    K_i,
                    M,
                    pivot_val,
                    pivot_val_dA,
                    pivot_val_dB,
                )

    return arr1_dA, arr1_dB


@njit
def fock_representation_diagonal_grad_NUMBA(
    A, B, M, cutoffs, arr0, arr2, arr1010, arr1001, arr1, tuple_type, list_type
):
    """
    Returns the gradients of the PNR probabilities of a mixed state according to algorithm 1 of
    https://doi.org/10.22331/q-2023-08-29-1097
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        arr0 (array): submatrix of the fock representation that contains Fock amplitudes of the type [a,a,b,b,c,c...]
        arr2 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+2,a,b,b,c,c...] / [a,a,b+2,b,c,c...] / ...
        arr1010 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b+1,b,c,c,...] / [a+1,a,b,b,c+1,c,...] / [a,a,b+1,b,c+1,c,...] / ...
        arr1001 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b+1,c,c,...] / [a+1,a,b,b,c,c+1,...] / [a,a,b+1,b,c,c+1,...] / ...
        arr1 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b,c,c...] / [a,a+1,b,b,c,c...] / [a,a,b+1,b,c,c...] / ...
        tuple_type, list_type (Numba types): numba types that need to be defined outside of Numba compiled functions
    Returns:
        array: the derivatives of the fock representation w.r.t. A and B
    """
    arr0_dA = np.zeros(arr0.shape + A.shape, dtype=np.complex128)
    arr2_dA = np.zeros(arr2.shape + A.shape, dtype=np.complex128)
    arr1010_dA = np.zeros(arr1010.shape + A.shape, dtype=np.complex128)
    arr1001_dA = np.zeros(arr1001.shape + A.shape, dtype=np.complex128)
    arr1_dA = np.zeros(arr1.shape + A.shape, dtype=np.complex128)
    arr0_dB = np.zeros(arr0.shape + B.shape, dtype=np.complex128)
    arr2_dB = np.zeros(arr2.shape + B.shape, dtype=np.complex128)
    arr1010_dB = np.zeros(arr1010.shape + B.shape, dtype=np.complex128)
    arr1001_dB = np.zeros(arr1001.shape + B.shape, dtype=np.complex128)
    arr1_dB = np.zeros(arr1.shape + B.shape, dtype=np.complex128)

    dict_params = construct_dict_params(cutoffs, tuple_type, list_type)
    for sum_params in range(sum(cutoffs)):
        for params in dict_params[sum_params]:
            # diagonal pivots: aa,bb,cc,dd,...
            if (cutoffs[0] == 1) or (params[0] < cutoffs[0] - 1):
                arr1_dA, arr1_dB = use_diag_pivot_grad(
                    A, B, M, cutoffs, params, arr0, arr1, arr0_dA, arr1_dA, arr0_dB, arr1_dB
                )
            # off-diagonal pivots: d=0: (a+1)a,bb,cc,dd,... | d=1: 00,(b+1)b,cc,dd | 00,00,(c+1)c,dd | ...
            for d in range(M):
                if np.all(np.array(params)[:d] == 0) and (params[d] < cutoffs[d] - 1):
                    (
                        arr0_dA,
                        arr2_dA,
                        arr1010_dA,
                        arr1001_dA,
                        arr0_dB,
                        arr2_dB,
                        arr1010_dB,
                        arr1001_dB,
                    ) = use_offDiag_pivot_grad(
                        A,
                        B,
                        M,
                        cutoffs,
                        params,
                        d,
                        arr0,
                        arr2,
                        arr1010,
                        arr1001,
                        arr1,
                        arr0_dA,
                        arr2_dA,
                        arr1010_dA,
                        arr1001_dA,
                        arr1_dA,
                        arr0_dB,
                        arr2_dB,
                        arr1010_dB,
                        arr1001_dB,
                        arr1_dB,
                    )
    return arr0_dA, arr0_dB


def fock_representation_diagonal_grad(A, B, M, arr0, arr2, arr1010, arr1001, arr1):
    """
    First initialise some Numba types (needs to be done outside of Numba compiled function)
    Then calculate the fock representation.
    """

    cutoffs = arr0.shape
    tuple_type = numba.types.UniTuple(int64, M)
    list_type = numba.types.ListType(tuple_type)
    return fock_representation_diagonal_grad_NUMBA(
        A, B, M, cutoffs, arr0, arr2, arr1010, arr1001, arr1, tuple_type, list_type
    )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from matplotlib import colors
import matplotlib.pyplot as plt
from mrmustard import math
from mrmustard.physics import bargmann
from mrmustard.physics.ansatze import Ansatz, PolyExpAnsatz
from mrmustard.utils.typing import Batch, ComplexMatrix, ComplexTensor, ComplexVector, Scalar
import numpy as np
from mrmustard import math


class Representation:
    r"""
    A base class for representations.
    """

    def from_ansatz(self, ansatz: Ansatz) -> Representation:
        r"""
        Returns a representation object from an ansatz object.
        To be implemented by subclasses.
        """
        raise NotImplementedError

    def __eq__(self, other: Representation) -> bool:
        r"""
        Whether this representation is equal to another.
        """
        return self.ansatz == other.ansatz

    def __add__(self, other: Representation) -> Representation:
        r"""
        Adds this representation to another.
        """
        return self.from_ansatz(self.ansatz + other.ansatz)

    def __sub__(self, other) -> Representation:
        r"""
        Subtracts another representation from this one.
        """
        return self.from_ansatz(self.ansatz - other.ansatz)

    def __mul__(self, other: Representation | Scalar) -> Representation:
        r"""
        Multiplies this representation by another or by a scalar.
        """
        try:
            return self.from_ansatz(self.ansatz * other.ansatz)
        except AttributeError:
            return self.from_ansatz(self.ansatz * other)

    def __rmul__(self, other: Representation | Scalar) -> Representation:
        r"""
        Multiplies this representation by another or by a scalar on the right.
        """
        return self.__mul__(other)

    def __truediv__(self, other: Representation | Scalar) -> Representation:
        r"""
        Divides this representation by another or by a scalar.
        """
        try:
            return self.from_ansatz(self.ansatz / other.ansatz)
        except AttributeError:
            return self.from_ansatz(self.ansatz / other)

    def __rtruediv__(self, other: Representation | Scalar) -> Representation:
        r"""
        Divides this representation by another or by a scalar on the right.
        """
        return self.from_ansatz(other / self.ansatz)

    def __and__(self, other: Representation) -> Representation:
        r"""
        Takes the outer product of this representation with another.
        """
        return self.from_ansatz(self.ansatz & other.ansatz)


class Bargmann(Representation):
    r"""This class is the Fock-Bargmann representation of a broad class of quantum states,
    transformations, measurements, channels, etc.

    The ansatz available in this representation is a linear combination of
    exponentials of bilinear forms with a polynomial part:

    .. math::
        F(z) = \sum_i \textrm{poly}_i(z) \textrm{exp}(z^T A_i z / 2 + z^T b_i)

    This function allows for vector space operations on Bargmann objects including
    linear combinations, outer product, and inner product.
    The inner product is defined as the contraction of two Bargmann objects across
    marked indices. This can also be used to contract existing indices
    in a single Bargmann object, e.g. to implement the partial trace.

    Note that the operations that change the shape of the ansatz (outer product (``&``)
    and inner product (``@``)) do not automatically modify the ordering of the
    combined or leftover indices.

    Examples:
        .. code-block:: python

            A = math.astensor([[[1.0]]])  # 1x1x1
            b = math.astensor([[0.0]])    # 1x1
            c = math.astensor([0.9])      # 1
            psi1 = Bargmann(A, b, c)
            psi2 = Bargmann(A, b, c)
            psi3 = 1.3 * psi1 - 2.1 * psi2  # linear combination
            assert psi3.A.shape == (2, 1, 1)  # stacked along batch dimension
            psi4 = psi1[0] @ psi2[0]  # contract wires 0 on each (inner product)
            assert psi4.A.shape == (1,)  # A is 0x0 now (no wires left)
            psi5 = psi1 & psi2  # outer product (tensor product)
            rho = psi1.conj() & psi1   # outer product (this is now the density matrix)
            assert rho.A.shape == (1, 2, 2)  # we have two wires now
            assert np.allclose(rho.trace((0,), (1,)), np.abs(c)**2)


    Args:
        A: batch of quadratic coefficient :math:`A_i`
        b: batch of linear coefficients :math:`b_i`
        c: batch of arrays :math:`c_i` (default: [1.0])
    """

    def __init__(
        self,
        A: Batch[ComplexMatrix],
        b: Batch[ComplexVector],
        c: Batch[ComplexTensor] = [1.0],
    ):
        r"""Initializes the Bargmann representation. Args can be passed non-batched,
        they will be automatically broadcasted to the correct batch shape.

        Args:
            A: batch of quadratic coefficient :math:`A_i`
            b: batch of linear coefficients :math:`b_i`
            c: batch of arrays :math:`c_i` (default: [1.0])
        """
        self._contract_idxs: tuple[int, ...] = ()
        self.ansatz = PolyExpAnsatz(A, b, c)

    def __call__(self, z: ComplexTensor) -> ComplexTensor:
        r"""Evaluates the Bargmann function at the given array of points."""
        return self.ansatz(z)

    def from_ansatz(self, ansatz: PolyExpAnsatz) -> Bargmann:
        r"""Returns a Bargmann object from an ansatz object."""
        return self.__class__(ansatz.A, ansatz.b, ansatz.c)

    @property
    def A(self) -> Batch[ComplexMatrix]:
        r"""
        The batch of quadratic coefficient :math:`A_i`.
        """
        return self.ansatz.A

    @property
    def b(self) -> Batch[ComplexVector]:
        r"""
        The batch of linear coefficients :math:`b_i`
        """
        return self.ansatz.b

    @property
    def c(self) -> Batch[ComplexTensor]:
        r"""
        The batch of arrays :math:`c_i`.
        """
        return self.ansatz.c

    def conj(self):
        r"""
        The conjugate of this Bargmann object.
        """
        new = self.__class__(math.conj(self.A), math.conj(self.b), math.conj(self.c))
        new._contract_idxs = self._contract_idxs
        return new

    def __getitem__(self, idx: int | tuple[int, ...]) -> Bargmann:
        r"""Returns a copy of self with the given indices marked for contraction."""
        idx = (idx,) if isinstance(idx, int) else idx
        for i in idx:
            if i >= self.ansatz.dim:
                raise IndexError(
                    f"Index {i} out of bounds for ansatz {self.ansatz.__class__.__qualname__} of dimension {self.ansatz.dim}."
                )
        new = self.__class__(self.A, self.b, self.c)
        new._contract_idxs = idx
        return new

    def __matmul__(self, other: Bargmann) -> Bargmann:
        r"""Implements the inner product of ansatze across the marked indices."""
        if self.ansatz.degree > 0 or other.ansatz.degree > 0:
            raise NotImplementedError(
                "Inner product of ansatze is only supported for ansatze with polynomial of degree 0."
            )
        Abc = []
        for A1, b1, c1 in zip(self.A, self.b, self.c):
            for A2, b2, c2 in zip(other.A, other.b, other.c):
                Abc.append(
                    bargmann.contract_two_Abc(
                        (A1, b1, c1),
                        (A2, b2, c2),
                        self._contract_idxs,
                        other._contract_idxs,
                    )
                )
        A, b, c = zip(*Abc)
        return self.__class__(math.astensor(A), math.astensor(b), math.astensor(c))

    def trace(self, idx_z: tuple[int, ...], idx_zconj: tuple[int, ...]) -> Bargmann:
        r"""Implements the partial trace over the given index pairs.

        Args:
            idx_z: indices to trace over
            idx_zconj: indices to trace over

        Returns:
            Bargmann: the ansatz with the given indices traced over
        """
        if self.ansatz.degree > 0:
            raise NotImplementedError(
                "Partial trace is only supported for ansatze with polynomial of degree ``0``."
            )
        if len(idx_z) != len(idx_zconj):
            msg = f"The number of indices to trace over must be the same for ``z`` and ``z*`` (got {len(idx_z)} and {len(idx_zconj)})."
            raise ValueError(msg)
        A, b, c = [], [], []
        for Abci in zip(self.A, self.b, self.c):
            Aij, bij, cij = bargmann.complex_gaussian_integral(Abci, idx_z, idx_zconj, measure=-1.0)
            A.append(Aij)
            b.append(bij)
            c.append(cij)
        return self.__class__(math.astensor(A), math.astensor(b), math.astensor(c))

    def reorder(self, order: tuple[int, ...] | list[int]) -> Bargmann:
        r"""Reorders the indices of the A matrix and b vector of an (A,b,c) triple.
        Returns a new Bargmann object."""
        A, b, c = bargmann.reorder_abc((self.A, self.b, self.c), order)
        return self.__class__(A, b, c)

    def plot(
        self,
        just_phase: bool = False,
        with_measure: bool = False,
        log_scale: bool = False,
        xlim=(-2 * np.pi, 2 * np.pi),
        ylim=(-2 * np.pi, 2 * np.pi),
        **kwargs,
    ):  # pragma: no cover
        r"""Plots the Bargmann function F(z) on the complex plane. Phase is represented by color,
        magnitude by brightness. The function can be multiplied by exp(-|z|^2) to represent
        the Bargmann function times the measure function (for integration).

        Args:
            just_phase (bool): whether to plot only the phase of the Bargmann function
            with_measure (bool): whether to plot the bargmann function times the measure function exp(-|z|^2)
            log_scale (bool): whether to plot the log of the Bargmann function
            xlim (tuple[float, float]): x limits of the plot
            ylim (tuple[float, float]): y limits of the plot

        Returns:
            tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]: the figure and axes of the plot
        """
        # eval F(z) on a grid of complex numbers
        X, Y = np.mgrid[xlim[0] : xlim[1] : 400j, ylim[0] : ylim[1] : 400j]
        Z = (X + 1j * Y).T
        f_values = self(Z[..., None])
        if log_scale:
            f_values = np.log(np.abs(f_values)) * np.exp(1j * np.angle(f_values))
        if with_measure:
            f_values = f_values * np.exp(-np.abs(Z) ** 2)

        # Get phase and magnitude of F(z)
        phases = np.angle(f_values) / (2 * np.pi) % 1
        magnitudes = np.abs(f_values)
        magnitudes_scaled = magnitudes / np.max(magnitudes)

        # Convert to RGB
        hsv_values = np.zeros(f_values.shape + (3,))
        hsv_values[..., 0] = phases
        hsv_values[..., 1] = 1
        hsv_values[..., 2] = 1 if just_phase else magnitudes_scaled
        rgb_values = colors.hsv_to_rgb(hsv_values)

        # Plot the image
        fig, ax = plt.subplots()
        ax.imshow(rgb_values, origin="lower", extent=[xlim[0], xlim[1], ylim[0], ylim[1]])
        ax.set_xlabel("$Re(z)$")
        ax.set_ylabel("$Im(z)$")

        name = "F_{" + self.ansatz.name + "}(z)"
        name = f"\\arg({name})\\log|{name}|" if log_scale else name
        title = name + "e^{-|z|^2}" if with_measure else name
        title = f"\\arg({name})" if just_phase else title
        ax.set_title(f"${title}$")
        plt.show(block=False)
        return fig, ax



# Copyright 2010 Pallets

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions for creating a logger that can be used in Mr Mustard.

The implementation in this module is based on the solution for logging used in
the Flask web application framework:
https://github.com/pallets/flask/blob/master/src/flask/logging.py
"""

import logging
import sys


def logging_handler_defined(logger):
    """Checks if the logger or any of its ancestors has a handler defined.

    The output depends on whether or not propagation was set for the logger.

    Args:
        logger (logging.Logger): the logger to check

    Returns:
        bool: whether or not a handler was defined
    """
    current = logger

    while current:
        if current.handlers:
            return True

        if not current.propagate:
            break

        current = current.parent

    return False


default_handler = logging.StreamHandler(sys.stderr)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
default_handler.setFormatter(formatter)


def create_logger(name, level=logging.INFO):
    """Get the Mr Mustard module specific logger and configure it if needed.

    Configuration only takes place if no user configuration was applied to the
    logger. Therefore, the logger is configured if and only if the following
    are true:

    - the logger has WARNING as effective level,
    - the level of the logger was not explicitly set,
    - no handlers were added to the logger.

    As the root logger has a WARNING level by default, any module specific
    logger will inherit the same as effective level.

    The default handler that is used for configuration writes to the standard
    error stream and uses a datetime and level formatter.

    Args:
        name (str): the name of the module for which the logger is being created
        level (logging.level): the logging level to set for the logger
    """
    logger = logging.getLogger(name)

    effective_level_inherited = logger.getEffectiveLevel() == logging.WARNING
    level_not_set = not logger.level
    no_handlers = not logging_handler_defined(logger)

    if effective_level_inherited and level_not_set and no_handlers:
        logger.setLevel(level)
        logger.addHandler(default_handler)

    return logger



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the backend manager."""


import importlib.util
import sys
from functools import lru_cache
from itertools import product
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import numpy as np
from scipy.special import binom
from scipy.stats import ortho_group, unitary_group

from ..utils.settings import settings
from ..utils.typing import (
    Batch,
    Matrix,
    Tensor,
    Trainable,
    Vector,
)
from .backend_base import BackendBase
from .backend_numpy import BackendNumpy

__all__ = [
    "BackendManager",
]

# ~~~~~~~
# Helpers
# ~~~~~~~


def lazy_import(module_name: str):
    r"""
    Returns module and loader for lazy import.

    Args:
        module_name: The name of the module to import.
    """
    try:
        return sys.modules[module_name], None
    except KeyError:
        spec = importlib.util.find_spec(module_name)
        module = importlib.util.module_from_spec(spec)
        loader = importlib.util.LazyLoader(spec.loader)
        return module, loader


# lazy import for numpy
module_name_np = "mrmustard.math.backend_numpy"
module_np, loader_np = lazy_import(module_name_np)

# lazy import for tensorflow
module_name_tf = "mrmustard.math.backend_tensorflow"
module_tf, loader_tf = lazy_import(module_name_tf)

all_modules = {
    "numpy": {"module": module_np, "loader": loader_np, "object": "BackendNumpy"},
    "tensorflow": {
        "module": module_tf,
        "loader": loader_tf,
        "object": "BackendTensorflow",
    },
}


class BackendManager:  # pylint: disable=too-many-public-methods, fixme
    r"""
    A class to manage the different backends supported by Mr Mustard.
    """

    # the backend in use, which is numpy by default
    _backend = BackendNumpy()

    # the configured Euclidean optimizer.
    _euclidean_opt: Optional[type] = None

    # whether or not the backend can be changed
    _is_immutable = False

    def __init__(self) -> None:
        # binding types and decorators of numpy backend
        self._bind()

    def _apply(self, fn: str, args: Optional[Sequence[Any]] = ()) -> Any:
        r"""
        Applies a function ``fn`` from the backend in use to the given ``args``.
        """
        try:
            attr = getattr(self.backend, fn)
        except AttributeError:
            msg = f"Function ``{fn}`` not implemented for backend ``{self.backend_name}``."
            # pylint: disable=raise-missing-from
            raise NotImplementedError(msg)
        return attr(*args)

    def _bind(self) -> None:
        r"""
        Binds the types and decorators of this backend manager to those of the given ``self._backend``.
        """
        for name in [
            "int32",
            "float32",
            "float64",
            "complex64",
            "complex128",
            "hermite_renormalized",
            "hermite_renormalized_binomial",
            "hermite_renormalized_diagonal_reorderedAB",
            "hermite_renormalized_1leftoverMode_reorderedAB",
        ]:
            setattr(self, name, getattr(self._backend, name))

    def __new__(cls):
        # singleton
        try:
            return cls.instance
        except AttributeError:
            cls.instance = super(BackendManager, cls).__new__(cls)
            return cls.instance

    def __repr__(self) -> str:
        return f"Backend({self.backend_name})"

    @property
    def backend(self) -> BackendBase:
        r"""
        The backend that is being used.
        """
        self._is_immutable = True
        return self._backend

    @property
    def backend_name(self) -> str:
        r"""
        The name of the backend in use.
        """
        return self._backend.name

    def change_backend(self, name: str) -> None:
        r"""
        Changes the backend to a different one.

        Args:
            name: The name of the new backend.
        """
        if name not in ["numpy", "tensorflow"]:
            msg = "Backend must be either ``numpy`` or ``tensorflow``"
            raise ValueError(msg)

        if self.backend_name != name:
            if self._is_immutable:
                msg = "Can no longer change the backend in this session."
                raise ValueError(msg)

            module = all_modules[name]["module"]
            object = all_modules[name]["object"]
            try:
                backend = getattr(module, object)()
            except AttributeError:
                # lazy import
                loader = all_modules[name]["loader"]
                loader.exec_module(module)
                backend = getattr(module, object)()

            # switch backend
            self._backend = backend

            # bind
            self._bind()

    # ~~~~~~~
    # Methods
    # ~~~~~~~
    # Below are the methods supported by the various backends.

    def abs(self, array: Tensor) -> Tensor:
        r"""The absolute value of array.

        Args:
            array: The array to take the absolute value of.

        Returns:
            The absolute value of the given ``array``.
        """
        return self._apply("abs", (array,))

    def any(self, array: Tensor) -> bool:
        r"""Returns ``True`` if any element of array is ``True``, ``False`` otherwise.

        Args:
            array: The array to check.

        Returns:
            ``True`` if any element of array is ``True``, ``False`` otherwise.
        """
        return self._apply("any", (array,))

    def arange(self, start: int, limit: int = None, delta: int = 1, dtype: Any = None) -> Tensor:
        r"""Returns an array of evenly spaced values within a given interval.

        Args:
            start: The start of the interval.
            limit: The end of the interval.
            delta: The step size.
            dtype: The dtype of the returned array.

        Returns:
            The array of evenly spaced values.
        """
        # NOTE: is float64 by default
        return self._apply("arange", (start, limit, delta, dtype))

    def asnumpy(self, tensor: Tensor) -> Tensor:
        r"""Converts an array to a numpy array.

        Args:
            tensor: The tensor to convert.

        Returns:
            The corresponidng numpy array.
        """
        return self._apply("asnumpy", (tensor,))

    def assign(self, tensor: Tensor, value: Tensor) -> Tensor:
        r"""Assigns value to tensor.

        Args:
            tensor: The tensor to assign to.
            value: The value to assign.

        Returns:
            The tensor with value assigned
        """
        return self._apply("assign", (tensor, value))

    def astensor(self, array: Tensor, dtype=None):
        r"""Converts a numpy array to a tensor.

        Args:
            array: The numpy array to convert.
            dtype: The dtype of the tensor.  If ``None``, the returned tensor
                is of type ``float``.

        Returns:
            The tensor with dtype.
        """
        return self._apply("astensor", (array, dtype))

    def atleast_1d(self, array: Tensor, dtype=None) -> Tensor:
        r"""Returns an array with at least one dimension.

        Args:
            array: The array to convert.
            dtype: The data type of the array. If ``None``, the returned array
                is of the same type as the given one.

        Returns:
            The array with at least one dimension.
        """
        return self._apply("atleast_1d", (array, dtype))

    def atleast_2d(self, array: Tensor, dtype=None) -> Tensor:
        r"""Returns an array with at least two dimensions.

        Args:
            array: The array to convert.
            dtype: The data type of the array. If ``None``, the returned array
                is of the same type as the given one.

        Returns:
            The array with at least two dimensions.
        """
        return self._apply("atleast_2d", (array, dtype))

    def atleast_3d(self, array: Tensor, dtype=None) -> Tensor:
        r"""Returns an array with at least three dimensions by eventually inserting
        new axes at the beginning. Note this is not the way atleast_3d works in numpy
        and tensorflow, where it adds at the beginning and/or end.

        Args:
            array: The array to convert.
            dtype: The data type of the array. If ``None``, the returned array
                is of the same type as the given one.

        Returns:
            The array with at least three dimensions.
        """
        return self._apply("atleast_3d", (array, dtype))

    def block_diag(self, mat1: Matrix, mat2: Matrix) -> Matrix:
        r"""Returns a block diagonal matrix from the given matrices.

        Args:
            mat1: A matrix.
            mat2: A matrix.

        Returns:
            A block diagonal matrix from the given matrices.
        """
        return self._apply("block_diag", (mat1, mat2))

    def boolean_mask(self, tensor: Tensor, mask: Tensor) -> Tensor:
        """
        Returns a tensor based on the truth value of the boolean mask.

        Args:
            tensor: A tensor.
            mask: A boolean mask.

        Returns:
            A tensor based on the truth value of the boolean mask.
        """
        return self._apply("boolean_mask", (tensor, mask))

    def block(self, blocks: List[List[Tensor]], axes=(-2, -1)) -> Tensor:
        r"""Returns a matrix made from the given blocks.

        Args:
            blocks: A list of lists of compatible blocks.
            axes: The axes to stack the blocks along.

        Returns:
            The matrix made of blocks.
        """
        return self._apply("block", (blocks, axes))

    def cast(self, array: Tensor, dtype=None) -> Tensor:
        r"""Casts ``array`` to ``dtype``.

        Args:
            array: The array to cast.
            dtype: The data type to cast to. If ``None``, the returned array
                is the same as the given one.

        Returns:
            The array cast to dtype.
        """
        return self._apply("cast", (array, dtype))

    def clip(self, array: Tensor, a_min: float, a_max: float) -> Tensor:
        r"""Clips array to the interval ``[a_min, a_max]``.

        Args:
            array: The array to clip.
            a_min: The minimum value.
            a_max: The maximum value.

        Returns:
            The clipped array.
        """
        return self._apply("clip", (array, a_min, a_max))

    def concat(self, values: Sequence[Tensor], axis: int) -> Tensor:
        r"""Concatenates values along the given axis.

        Args:
            values: The values to concatenate.
            axis: The axis along which to concatenate.

        Returns:
            The concatenated values.
        """
        return self._apply("concat", (values, axis))

    def conj(self, array: Tensor) -> Tensor:
        r"""The complex conjugate of array.

        Args:
            array: The array to take the complex conjugate of.

        Returns:
            The complex conjugate of the given ``array``.
        """
        return self._apply("conj", (array,))

    def constraint_func(
        self, bounds: Tuple[Optional[float], Optional[float]]
    ) -> Optional[Callable]:
        r"""Returns a constraint function for the given bounds.

        A constraint function will clip the value to the interval given by the bounds.

        .. note::

            The upper and/or lower bounds can be ``None``, in which case the constraint
            function will not clip the value.

        Args:
            bounds: The bounds of the constraint.

        Returns:
            The constraint function.
        """
        return self._apply("constraint_func", (bounds))

    def convolution(
        self,
        array: Tensor,
        filters: Tensor,
        padding: Optional[str] = None,
        data_format="NWC",
    ) -> Tensor:  # TODO: remove strides and data_format?
        r"""Performs a convolution on array with filters.

        Args:
            array: The array to convolve.
            filters: The filters to convolve with.
            padding: The padding mode.
            data_format: The data format of the array.

        Returns:
            The convolved array.
        """
        return self._apply("convolution", (array, filters, padding, data_format))

    def cos(self, array: Tensor) -> Tensor:
        r"""The cosine of an array.

        Args:
            array: The array to take the cosine of.

        Returns:
            The cosine of ``array``.
        """
        return self._apply("cos", (array,))

    def cosh(self, array: Tensor) -> Tensor:
        r"""The hyperbolic cosine of array.

        Args:
            array: The array to take the hyperbolic cosine of.

        Returns:
            The hyperbolic cosine of ``array``.
        """
        return self._apply("cosh", (array,))

    def det(self, matrix: Tensor) -> Tensor:
        r"""The determinant of matrix.

        Args:
            matrix: The matrix to take the determinant of

        Returns:
            The determinant of ``matrix``.
        """
        return self._apply("det", (matrix,))

    def diag(self, array: Tensor, k: int = 0) -> Tensor:
        r"""The array made by inserting the given array along the :math:`k`-th diagonal.

        Args:
            array: The array to insert.
            k: The ``k``-th diagonal to insert array into.

        Returns:
            The array with ``array`` inserted into the ``k``-th diagonal.
        """
        return self._apply("diag", (array, k))

    def diag_part(self, array: Tensor, k: int = 0) -> Tensor:
        r"""The array of the main diagonal of array.

        Args:
            array: The array to extract the main diagonal of.
            k: The diagonal to extract.

        Returns:
            The array of the main diagonal of ``array``.
        """
        return self._apply("diag_part", (array, k))

    def eigvals(self, tensor: Tensor) -> Tensor:
        r"""The eigenvalues of a tensor.

        Args:
            tensor: The tensor to calculate the eigenvalues of.

        Returns:
            The eigenvalues of ``tensor``.
        """
        return self._apply("eigvals", (tensor,))

    def eigh(self, tensor: Tensor) -> Tensor:
        """
        The eigenvalues and eigenvectors of a matrix.

        Args:
            tensor: The tensor to calculate the eigenvalues and eigenvectors of.

        Returns:
            The eigenvalues and eigenvectors of ``tensor``.
        """
        return self._apply("eigh", (tensor,))

    def einsum(self, string: str, *tensors) -> Tensor:
        r"""The result of the Einstein summation convention on the tensors.

        Args:
            string: The string of the Einstein summation convention.
            tensors: The tensors to perform the Einstein summation on.

        Returns:
            The result of the Einstein summation convention.
        """
        return self._apply("einsum", (string, *tensors))

    def exp(self, array: Tensor) -> Tensor:
        r"""The exponential of array element-wise.

        Args:
            array: The array to take the exponential of.

        Returns:
            The exponential of array.
        """
        return self._apply("exp", (array,))

    def expand_dims(self, array: Tensor, axis: int) -> Tensor:
        r"""The array with an additional dimension inserted at the given axis.

        Args:
            array: The array to expand.
            axis: The axis to insert the new dimension.

        Returns:
            The array with an additional dimension inserted at the given axis.
        """
        return self._apply("expand_dims", (array, axis))

    def expm(self, matrix: Tensor) -> Tensor:
        r"""The matrix exponential of matrix.

        Args:
            matrix: The matrix to take the exponential of.

        Returns:
            The exponential of ``matrix``.
        """
        return self._apply("expm", (matrix,))

    def eye(self, size: int, dtype=None) -> Tensor:
        r"""The identity matrix of size.

        Args:
            size: The size of the identity matrix
            dtype: The data type of the identity matrix. If ``None``,
                the returned matrix is of type ``float``.

        Returns:
            The identity matrix.
        """
        return self._apply("eye", (size, dtype))

    def eye_like(self, array: Tensor) -> Tensor:
        r"""The identity matrix of the same shape and dtype as array.

        Args:
            array: The array to create the identity matrix of.

        Returns:
            The identity matrix.
        """
        return self._apply("eye_like", (array,))

    def from_backend(self, value: Any) -> bool:
        r"""Whether the given tensor is a tensor of the concrete backend.

        Args:
            value: A value.

        Returns:
            Whether given ``value`` is a tensor of the concrete backend.
        """
        return self._apply("from_backend", (value,))

    def gather(self, array: Tensor, indices: Batch[int], axis: Optional[int] = None) -> Tensor:
        r"""The values of the array at the given indices.

        Args:
            array: The array to gather values from.
            indices: The indices to gather values from.
            axis: The axis to gather values from.

        Returns:
            The values of the array at the given indices.
        """
        return self._apply(
            "gather",
            (
                array,
                indices,
                axis,
            ),
        )

    def hermite_renormalized_batch(
        self, A: Tensor, B: Tensor, C: Tensor, shape: Tuple[int]
    ) -> Tensor:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx + 1/2*Ax^2)` at zero, where the series has :math:`sqrt(n!)`
        at the denominator rather than :math:`n!`. It computes all the amplitudes within the
        tensor of given shape in case of B is a batched vector with a batched diemnsion on the
        last index.

        Args:
            A: The A matrix.
            B: The batched B vector with its batch dimension on the last index.
            C: The C scalar.
            shape: The shape of the final tensor.

        Returns:
            The batched Hermite polynomial of given shape.
        """
        return self._apply("hermite_renormalized_batch", (A, B, C, shape))

    def hermite_renormalized_diagonal(
        self, A: Tensor, B: Tensor, C: Tensor, cutoffs: Tuple[int]
    ) -> Tensor:
        r"""Firsts, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.compactFock~
        Then, calculates the required renormalized multidimensional Hermite polynomial.
        """
        return self._apply("hermite_renormalized_diagonal", (A, B, C, cutoffs))

    def hermite_renormalized_diagonal_batch(
        self, A: Tensor, B: Tensor, C: Tensor, cutoffs: Tuple[int]
    ) -> Tensor:
        r"""First, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.compactFock~
        Then, calculates the required renormalized multidimensional Hermite polynomial.
        Same as hermite_renormalized_diagonal but works for a batch of different B's."""
        return self._apply("hermite_renormalized_diagonal_batch", (A, B, C, cutoffs))

    def hermite_renormalized_1leftoverMode(
        self, A: Tensor, B: Tensor, C: Tensor, cutoffs: Tuple[int]
    ) -> Tensor:
        r"""First, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.compactFock~
        Then, calculate the required renormalized multidimensional Hermite polynomial.
        """
        return self._apply("hermite_renormalized_1leftoverMode", (A, B, C, cutoffs))

    def imag(self, array: Tensor) -> Tensor:
        r"""The imaginary part of array.

        Args:
            array: The array to take the imaginary part of

        Returns:
            The imaginary part of array
        """
        return self._apply("imag", (array,))

    def inv(self, tensor: Tensor) -> Tensor:
        r"""The inverse of tensor.

        Args:
            tensor: The tensor to take the inverse of

        Returns:
            The inverse of tensor
        """
        return self._apply("inv", (tensor,))

    def is_trainable(self, tensor: Tensor) -> bool:
        r"""Whether the given tensor is trainable.

        Args:
            tensor: The tensor to train.

        Returns:
            Whether the given tensor can be trained.
        """
        return self._apply("is_trainable", (tensor,))

    def lgamma(self, x: Tensor) -> Tensor:
        r"""The natural logarithm of the gamma function of ``x``.

        Args:
            x: The array to take the natural logarithm of the gamma function of

        Returns:
            The natural logarithm of the gamma function of ``x``
        """
        return self._apply("lgamma", (x,))

    def log(self, x: Tensor) -> Tensor:
        r"""The natural logarithm of ``x``.

        Args:
            x: The array to take the natural logarithm of

        Returns:
            The natural logarithm of ``x``
        """
        return self._apply("log", (x,))

    def make_complex(self, real: Tensor, imag: Tensor) -> Tensor:
        """Given two real tensors representing the real and imaginary part of a complex number,
        this operation returns a complex tensor. The input tensors must have the same shape.

        Args:
            real: The real part of the complex number.
            imag: The imaginary part of the complex number.

        Returns:
            The complex array ``real + 1j * imag``.
        """
        return self._apply("make_complex", (real, imag))

    def matmul(self, *matrices: Matrix) -> Tensor:
        r"""The matrix product of the given matrices.

        Args:
            matrices: The matrices to multiply.

        Returns:
            The matrix product
        """
        return self._apply("matmul", matrices)

    def matvec(self, a: Matrix, b: Vector) -> Tensor:
        r"""The matrix vector product of ``a`` (matrix) and ``b`` (vector).

        Args:
            a: The matrix to multiply
            b: The vector to multiply

        Returns:
            The matrix vector product of ``a`` and ``b``
        """
        return self._apply("matvec", (a, b))

    def maximum(self, a: Tensor, b: Tensor) -> Tensor:
        r"""The element-wise maximum of ``a`` and ``b``.

        Args:
            a: The first array to take the maximum of
            b: The second array to take the maximum of

        Returns:
            The element-wise maximum of ``a`` and ``b``
        """
        return self._apply(
            "maximum",
            (
                a,
                b,
            ),
        )

    def minimum(self, a: Tensor, b: Tensor) -> Tensor:
        r"""The element-wise minimum of ``a`` and ``b``.

        Args:
            a: The first array to take the minimum of
            b: The second array to take the minimum of

        Returns:
            The element-wise minimum of ``a`` and ``b``
        """
        return self._apply(
            "minimum",
            (
                a,
                b,
            ),
        )

    def new_variable(
        self,
        value: Tensor,
        bounds: Tuple[Optional[float], Optional[float]],
        name: str,
        dtype=None,
    ) -> Tensor:
        r"""Returns a new variable with the given value and bounds.

        Args:
            value: The value of the new variable.
            bounds: The bounds of the new variable.
            name: The name of the new variable.
            dtype: dtype of the new variable. If ``None``, casts it to float.
        Returns:
            The new variable.
        """
        return self._apply("new_variable", (value, bounds, name, dtype))

    def new_constant(self, value: Tensor, name: str, dtype=None) -> Tensor:
        r"""Returns a new constant with the given value.

        Args:
            value: The value of the new constant
            name (str): name of the new constant
            dtype (type): dtype of the array

        Returns:
            The new constant
        """
        return self._apply("new_constant", (value, name, dtype))

    def norm(self, array: Tensor) -> Tensor:
        r"""The norm of array.

        Args:
            array: The array to take the norm of

        Returns:
            The norm of array
        """
        return self._apply("norm", (array,))

    def ones(self, shape: Sequence[int], dtype=None) -> Tensor:
        r"""Returns an array of ones with the given ``shape`` and ``dtype``.

        Args:
            shape (tuple): shape of the array
            dtype (type): dtype of the array. If ``None``, the returned array is
                of type ``float``.

        Returns:
            The array of ones
        """
        # NOTE : should be float64 by default
        return self._apply("ones", (shape, dtype))

    def ones_like(self, array: Tensor) -> Tensor:
        r"""Returns an array of ones with the same shape and ``dtype`` as ``array``.

        Args:
            array: The array to take the shape and dtype of

        Returns:
            The array of ones
        """
        return self._apply("ones_like", (array,))

    def outer(self, array1: Tensor, array2: Tensor) -> Tensor:
        r"""The outer product of ``array1`` and ``array2``.

        Args:
            array1: The first array to take the outer product of
            array2: The second array to take the outer product of

        Returns:
            The outer product of array1 and array2
        """
        return self._apply("outer", (array1, array2))

    def pad(
        self,
        array: Tensor,
        paddings: Sequence[Tuple[int, int]],
        mode="CONSTANT",
        constant_values=0,
    ) -> Tensor:
        r"""The padded array.

        Args:
            array: The array to pad
            paddings (tuple): paddings to apply
            mode (str): mode to apply the padding
            constant_values (int): constant values to use for padding

        Returns:
            The padded array
        """
        return self._apply("pad", (array, paddings, mode, constant_values))

    def pinv(self, matrix: Tensor) -> Tensor:
        r"""The pseudo-inverse of matrix.

        Args:
            matrix: The matrix to take the pseudo-inverse of

        Returns:
            The pseudo-inverse of matrix
        """
        return self._apply("pinv", (matrix,))

    def pow(self, x: Tensor, y: Tensor) -> Tensor:
        r"""Returns :math:`x^y`. Broadcasts ``x`` and ``y`` if necessary.
        Args:
            x: The base
            y: The exponent

        Returns:
            The :math:`x^y`
        """
        return self._apply("pow", (x, y))

    def real(self, array: Tensor) -> Tensor:
        r"""The real part of ``array``.

        Args:
            array: The array to take the real part of

        Returns:
            The real part of ``array``
        """
        return self._apply("real", (array,))

    def reshape(self, array: Tensor, shape: Sequence[int]) -> Tensor:
        r"""The reshaped array.

        Args:
            array: The array to reshape
            shape (tuple): shape to reshape the array to

        Returns:
            The reshaped array
        """
        return self._apply("reshape", (array, shape))

    def round(self, array: Tensor, decimals: int) -> Tensor:
        r"""The array rounded to the nearest integer.

        Args:
            array: The array to round
            decimals: number of decimals to round to

        Returns:
            The array rounded to the nearest integer
        """
        return self._apply("round", (array, decimals))

    def set_diag(self, array: Tensor, diag: Tensor, k: int) -> Tensor:
        r"""The array with the diagonal set to ``diag``.

        Args:
            array: The array to set the diagonal of
            diag: The diagonal to set
            k (int): diagonal to set

        Returns:
            The array with the diagonal set to ``diag``
        """
        return self._apply("set_diag", (array, diag, k))

    def sin(self, array: Tensor) -> Tensor:
        r"""The sine of ``array``.

        Args:
            array: The array to take the sine of

        Returns:
            The sine of ``array``
        """
        return self._apply("sin", (array,))

    def sinh(self, array: Tensor) -> Tensor:
        r"""The hyperbolic sine of ``array``.

        Args:
            array: The array to take the hyperbolic sine of

        Returns:
            The hyperbolic sine of ``array``
        """
        return self._apply("sinh", (array,))

    def solve(self, matrix: Tensor, rhs: Tensor) -> Tensor:
        r"""The solution of the linear system :math:`Ax = b`.

        Args:
            matrix: The matrix :math:`A`
            rhs: The vector :math:`b`

        Returns:
            The solution :math:`x`
        """
        return self._apply("solve", (matrix, rhs))

    def sqrt(self, x: Tensor, dtype=None) -> Tensor:
        r"""The square root of ``x``.

        Args:
            x: The array to take the square root of
            dtype: ``dtype`` of the output array.

        Returns:
            The square root of ``x``
        """
        return self._apply("sqrt", (x, dtype))

    def sqrtm(self, tensor: Tensor, dtype=None) -> Tensor:
        r"""The matrix square root.

        Args:
            tensor: The tensor to take the matrix square root of.
            dtype: The ``dtype`` of the output tensor. If ``None``, the output
                is of type ``math.complex128``.

        Returns:
            The square root of ``x``"""
        return self._apply("sqrtm", (tensor, dtype))

    def sum(self, array: Tensor, axes: Sequence[int] = None):
        r"""The sum of array.

        Args:
            array: The array to take the sum of
            axes (tuple): axes to sum over

        Returns:
            The sum of array
        """
        if axes is not None:
            neg = [a for a in axes if a < 0]
            pos = [a for a in axes if a >= 0]
            axes = sorted(neg) + sorted(pos)[::-1]
        return self._apply("sum", (array, axes))

    def tensordot(self, a: Tensor, b: Tensor, axes: Sequence[int]) -> Tensor:
        r"""The tensordot product of ``a`` and ``b``.

        Args:
            a: The first array to take the tensordot product of
            b: The second array to take the tensordot product of
            axes: The axes to take the tensordot product over

        Returns:
            The tensordot product of ``a`` and ``b``
        """
        return self._apply("tensordot", (a, b, axes))

    def tile(self, array: Tensor, repeats: Sequence[int]) -> Tensor:
        r"""The tiled array.

        Args:
            array: The array to tile
            repeats (tuple): number of times to tile the array along each axis

        Returns:
            The tiled array
        """
        return self._apply("tile", (array, repeats))

    def trace(self, array: Tensor, dtype=None) -> Tensor:
        r"""The trace of array.

        Args:
            array: The array to take the trace of
            dtype (type): ``dtype`` of the output array

        Returns:
            The trace of array
        """
        return self._apply("trace", (array, dtype))

    def transpose(self, a: Tensor, perm: Sequence[int] = None):
        r"""The transposed arrays.

        Args:
            a: The array to transpose
            perm (tuple): permutation to apply to the array

        Returns:
            The transposed array
        """
        return self._apply("transpose", (a, perm))

    def update_tensor(self, tensor: Tensor, indices: Tensor, values: Tensor) -> Tensor:
        r"""Updates a tensor in place with the given values.

        Args:
            tensor: The tensor to update
            indices: The indices to update
            values: The values to update

        Returns:
            The updated tensor
        """
        return self._apply("update_tensor", (tensor, indices, values))

    def update_add_tensor(self, tensor: Tensor, indices: Tensor, values: Tensor) -> Tensor:
        r"""Updates a tensor in place by adding the given values.

        Args:
            tensor: The tensor to update
            indices: The indices to update
            values: The values to add

        Returns:
            The updated tensor
        """
        return self._apply("update_add_tensor", (tensor, indices, values))

    def value_and_gradients(
        self, cost_fn: Callable, parameters: Dict[str, List[Trainable]]
    ) -> Tuple[Tensor, Dict[str, List[Tensor]]]:
        r"""The loss and gradients of the given cost function.

        Args:
            cost_fn (callable): cost function to compute the loss and gradients of
            parameters (dict): parameters to compute the loss and gradients of

        Returns:
            tuple: loss and gradients (dict) of the given cost function
        """
        return self._apply("value_and_gradients", (cost_fn, parameters))

    def xlogy(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Returns ``0`` if ``x == 0`` elementwise and ``x * log(y)`` otherwise.
        """
        return self._apply("xlogy", (x, y))

    def zeros(self, shape: Sequence[int], dtype=None) -> Tensor:
        r"""Returns an array of zeros with the given shape and ``dtype``.

        Args:
            shape: The shape of the array.
            dtype: The dtype of the array. If ``None``, the returned array is
                of type ``float``.

        Returns:
            The array of zeros.
        """
        return self._apply("zeros", (shape, dtype))

    def zeros_like(self, array: Tensor) -> Tensor:
        r"""Returns an array of zeros with the same shape and ``dtype`` as ``array``.

        Args:
            array: The array to take the shape and ``dtype`` of.

        Returns:
            The array of zeros.
        """
        return self._apply("zeros_like", (array,))

    def map_fn(self, fn: Callable, elements: Tensor) -> Tensor:
        """Transforms elems by applying fn to each element unstacked on axis 0.

        Args:
            fn (func): The callable to be performed. It accepts one argument,
                which will have the same (possibly nested) structure as elems.
            elements (Tensor): A tensor or (possibly nested) sequence of tensors,
                each of which will be unstacked along their first dimension.
                ``func`` will be applied to the nested sequence of the resulting slices.

        Returns:
            Tensor: applied ``func`` on ``elements``
        """
        return self._apply("map_fn", (fn, elements))

    def squeeze(self, tensor: Tensor, axis: Optional[List[int]]) -> Tensor:
        """Removes dimensions of size 1 from the shape of a tensor.

        Args:
            tensor (Tensor): the tensor to squeeze
            axis (Optional[List[int]]): if specified, only squeezes the
                dimensions listed, defaults to []

        Returns:
            Tensor: tensor with one or more dimensions of size 1 removed
        """
        return self._apply("squeeze", (tensor, axis))

    def cholesky(self, input: Tensor) -> Tensor:
        """Computes the Cholesky decomposition of square matrices.

        Args:
            input (Tensor)

        Returns:
            Tensor: tensor with the same type as input
        """
        return self._apply("cholesky", (input,))

    def Categorical(self, probs: Tensor, name: str):
        """Categorical distribution over integers.

        Args:
            probs: The unnormalized probabilities of a set of Categorical distributions.
            name: The name prefixed to operations created by this class.

        Returns:
            tfp.distributions.Categorical: instance of ``tfp.distributions.Categorical`` class
        """
        return self._apply("Categorical", (probs, name))

    def MultivariateNormalTriL(self, loc: Tensor, scale_tril: Tensor):
        """Multivariate normal distribution on `R^k` and parameterized by a (batch of) length-k loc
        vector (aka "mu") and a (batch of) k x k scale matrix; covariance = scale @ scale.T
        where @ denotes matrix-multiplication.

        Args:
            loc (Tensor): if this is set to None, loc is implicitly 0
            scale_tril: lower-triangular Tensor with non-zero diagonal elements

        Returns:
            tfp.distributions.MultivariateNormalTriL: instance of ``tfp.distributions.MultivariateNormalTriL``
        """
        return self._apply("MultivariateNormalTriL", (loc, scale_tril))

    def custom_gradient(self, func):
        r"""
        A decorator to define a function with a custom gradient.
        """

        def wrapper(*args, **kwargs):
            if self.backend_name == "numpy":
                return func(*args, **kwargs)
            else:
                from tensorflow import custom_gradient  # pylint: disable=import-outside-toplevel

                return custom_gradient(func)(*args, **kwargs)

        return wrapper

    def DefaultEuclideanOptimizer(self):
        r"""Default optimizer for the Euclidean parameters."""
        return self._apply("DefaultEuclideanOptimizer")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Methods that build on the basic ops and don't need to be overridden in the backend implementation
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    @property
    def euclidean_opt(self):
        r"""The configured Euclidean optimizer."""
        if not self._euclidean_opt:
            self._euclidean_opt = self.DefaultEuclideanOptimizer()
        return self._euclidean_opt

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Methods that build on the basic ops and don't need to be overridden in the backend implementation
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def dagger(self, array: Tensor) -> Tensor:
        """The adjoint of ``array``. This operation swaps the first
        and second half of the indexes and then conjugates the matrix.

        Args:
            array: The array to take the adjoint of

        Returns:
            The adjoint of ``array``
        """
        N = len(array.shape) // 2
        perm = list(range(N, 2 * N)) + list(range(0, N))
        return self.conj(self.transpose(array, perm=perm))

    def unitary_to_orthogonal(self, U):
        r"""Unitary to orthogonal mapping.

        Args:
            U: The unitary matrix in ``U(n)``

        Returns:
            The orthogonal matrix in :math:`O(2n)`
        """
        X = self.real(U)
        Y = self.imag(U)
        return self.block([[X, -Y], [Y, X]])

    def random_symplectic(self, num_modes: int, max_r: float = 1.0) -> Tensor:
        r"""A random symplectic matrix in ``Sp(2*num_modes)``.

        Squeezing is sampled uniformly from 0.0 to ``max_r`` (1.0 by default).
        """
        if num_modes == 1:
            W = np.exp(1j * settings.rng.uniform(size=(1, 1)))
            V = np.exp(1j * settings.rng.uniform(size=(1, 1)))
        else:
            W = unitary_group.rvs(dim=num_modes, random_state=settings.rng)
            V = unitary_group.rvs(dim=num_modes, random_state=settings.rng)
        r = settings.rng.uniform(low=0.0, high=max_r, size=num_modes)
        OW = self.unitary_to_orthogonal(W)
        OV = self.unitary_to_orthogonal(V)
        dd = self.diag(self.concat([self.exp(-r), np.exp(r)], axis=0), k=0)
        return OW @ dd @ OV

    @staticmethod
    def random_orthogonal(N: int) -> Tensor:
        """A random orthogonal matrix in :math:`O(N)`."""
        if N == 1:
            return np.array([[1.0]])
        return ortho_group.rvs(dim=N, random_state=settings.rng)

    def random_unitary(self, N: int) -> Tensor:
        """a random unitary matrix in :math:`U(N)`"""
        if N == 1:
            return self.exp(1j * settings.rng.uniform(size=(1, 1)))
        return unitary_group.rvs(dim=N, random_state=settings.rng)

    def single_mode_to_multimode_vec(self, vec, num_modes: int):
        r"""Apply the same 2-vector (i.e. single-mode) to a larger number of modes."""
        if vec.shape[-1] != 2:
            raise ValueError("vec must be 2-dimensional (i.e. single-mode)")
        x, y = vec[..., -2], vec[..., -1]
        vec = self.concat([self.tile([x], [num_modes]), self.tile([y], [num_modes])], axis=-1)
        return vec

    def single_mode_to_multimode_mat(self, mat: Tensor, num_modes: int):
        r"""Apply the same :math:`2\times 2` matrix (i.e. single-mode) to a larger number of modes."""
        if mat.shape[-2:] != (2, 2):
            raise ValueError("mat must be a single-mode (2x2) matrix")
        mat = self.diag(
            self.tile(self.expand_dims(mat, axis=-1), (1, 1, num_modes)), k=0
        )  # shape [2,2,N,N]
        mat = self.reshape(self.transpose(mat, (0, 2, 1, 3)), [2 * num_modes, 2 * num_modes])
        return mat

    @staticmethod
    @lru_cache()
    def Xmat(num_modes: int):
        r"""The matrix :math:`X_n = \begin{bmatrix}0 & I_n\\ I_n & 0\end{bmatrix}.`

        Args:
            num_modes (int): positive integer

        Returns:
            The :math:`2N\times 2N` array
        """
        I = np.identity(num_modes)
        O = np.zeros((num_modes, num_modes))
        return np.block([[O, I], [I, O]])

    @staticmethod
    @lru_cache()
    def rotmat(num_modes: int):
        "Rotation matrix from quadratures to complex amplitudes."
        I = np.identity(num_modes)
        return np.sqrt(0.5) * np.block([[I, 1j * I], [I, -1j * I]])

    @staticmethod
    @lru_cache()
    def J(num_modes: int):
        """Symplectic form."""
        I = np.identity(num_modes)
        O = np.zeros_like(I)
        return np.block([[O, I], [-I, O]])

    def add_at_modes(
        self, old: Tensor, new: Optional[Tensor], modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        """Adds two phase-space tensors (cov matrices, displacement vectors, etc..) on the specified modes."""
        if new is None:
            return old
        shape = getattr(old, "shape", ())
        N = (shape[-1] if shape != () else 0) // 2
        indices = modes + [m + N for m in modes]
        return self.update_add_tensor(
            old, list(product(*[indices] * len(new.shape))), self.reshape(new, -1)
        )

    def left_matmul_at_modes(
        self, a_partial: Tensor, b_full: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        r"""Left matrix multiplication of a partial matrix and a full matrix.

        It assumes that that ``a_partial`` is a matrix operating on M modes and that ``modes`` is a
        list of ``M`` integers, i.e., it will apply ``a_partial`` on the corresponding ``M`` modes
        of ``b_full`` from the left.

        Args:
            a_partial: The :math:`2M\times 2M` array
            b_full: The :math:`2N\times 2N` array
            modes: A list of ``M`` modes to perform the multiplication on

        Returns:
            The :math:`2N\times 2N` array
        """
        if a_partial is None:
            return b_full

        N = b_full.shape[-1] // 2
        indices = self.astensor(modes + [m + N for m in modes], dtype="int32")
        b_rows = self.gather(b_full, indices, axis=0)
        b_rows = self.matmul(a_partial, b_rows)
        return self.update_tensor(b_full, indices[:, None], b_rows)

    def right_matmul_at_modes(
        self, a_full: Tensor, b_partial: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        r"""Right matrix multiplication of a full matrix and a partial matrix.

        It assumes that that ``b_partial`` is a matrix operating on ``M`` modes and that ``modes``
        is a list of ``M`` integers, i.e., it will apply ``b_partial`` on the corresponding M modes
        of ``a_full`` from the right.

        Args:
            a_full: The :math:`2N\times 2N` array
            b_partial: The :math:`2M\times 2M` array
            modes: A list of `M` modes to perform the multiplication on

        Returns:
            The :math:`2N\times 2N` array
        """
        return self.transpose(
            self.left_matmul_at_modes(self.transpose(b_partial), self.transpose(a_full), modes)
        )

    def matvec_at_modes(
        self, mat: Optional[Tensor], vec: Tensor, modes: Sequence[int]
    ) -> Tensor:  # NOTE: To be deprecated (XPTensor)
        """Matrix-vector multiplication between a phase-space matrix and a vector in the specified modes."""
        if mat is None:
            return vec
        N = vec.shape[-1] // 2
        indices = self.astensor(modes + [m + N for m in modes], dtype="int32")
        updates = self.matvec(mat, self.gather(vec, indices, axis=0))
        return self.update_tensor(vec, indices[:, None], updates)

    def all_diagonals(self, rho: Tensor, real: bool) -> Tensor:
        """Returns all the diagonals of a density matrix."""
        cutoffs = rho.shape[: rho.ndim // 2]
        rho = self.reshape(rho, (int(np.prod(cutoffs)), int(np.prod(cutoffs))))
        diag = self.diag_part(rho)
        if real:
            return self.real(self.reshape(diag, cutoffs))

        return self.reshape(diag, cutoffs)

    def poisson(self, max_k: int, rate: Tensor) -> Tensor:
        """Poisson distribution up to ``max_k``."""
        k = self.arange(max_k)
        rate = self.cast(rate, k.dtype)
        return self.exp(k * self.log(rate + 1e-9) - rate - self.lgamma(k + 1.0))

    def binomial_conditional_prob(self, success_prob: Tensor, dim_out: int, dim_in: int):
        """:math:`P(out|in) = binom(in, out) * (1-success_prob)**(in-out) * success_prob**out`."""
        in_ = self.arange(dim_in)[None, :]
        out_ = self.arange(dim_out)[:, None]
        return (
            self.cast(binom(in_, out_), in_.dtype)
            * self.pow(success_prob, out_)
            * self.pow(1.0 - success_prob, self.maximum(in_ - out_, 0.0))
        )

    def convolve_probs_1d(self, prob: Tensor, other_probs: List[Tensor]) -> Tensor:
        """Convolution of a joint probability with a list of single-index probabilities."""

        if prob.ndim > 3 or len(other_probs) > 3:
            raise ValueError("cannot convolve arrays with more than 3 axes")
        if not all((q.ndim == 1 for q in other_probs)):
            raise ValueError("other_probs must contain 1d arrays")
        if not all((len(q) == s for q, s in zip(other_probs, prob.shape))):
            raise ValueError("The length of the 1d prob vectors must match shape of prob")

        q = other_probs[0]
        for q_ in other_probs[1:]:
            q = q[..., None] * q_[(None,) * q.ndim + (slice(None),)]

        return self.convolve_probs(prob, q)

    def convolve_probs(self, prob: Tensor, other: Tensor) -> Tensor:
        r"""Convolve two probability distributions (up to 3D) with the same shape.

        Note that the output is not guaranteed to be a complete joint probability,
        as it's computed only up to the dimension of the base probs.
        """
        if prob.ndim > 3 or other.ndim > 3:
            raise ValueError("cannot convolve arrays with more than 3 axes")
        if not prob.shape == other.shape:
            raise ValueError("prob and other must have the same shape")

        prob_padded = self.pad(prob, [(s - 1, 0) for s in other.shape])
        other_reversed = other[(slice(None, None, -1),) * other.ndim]
        return self.convolution(
            prob_padded[None, ..., None],
            other_reversed[..., None, None],
            data_format="N"
            + ("HD"[: other.ndim - 1])[::-1]
            + "WC",  # TODO: rewrite this to be more readable (do we need it?)
        )[0, ..., 0]

    def euclidean_to_symplectic(self, S: Matrix, dS_euclidean: Matrix) -> Matrix:
        r"""Convert the Euclidean gradient to a Riemannian gradient on the
        tangent bundle of the symplectic manifold.

        Implemented from:
            Wang J, Sun H, Fiori S. A Riemannian‐steepest‐descent approach
            for optimization on the real symplectic group.
            Mathematical Methods in the Applied Sciences. 2018 Jul 30;41(11):4273-86.

        Args:
            S (Matrix): symplectic matrix
            dS_euclidean (Matrix): Euclidean gradient tensor

        Returns:
            Matrix: symplectic gradient tensor
        """
        Jmat = self.J(S.shape[-1] // 2)
        Z = self.matmul(self.transpose(S), dS_euclidean)
        return 0.5 * (Z + self.matmul(self.matmul(Jmat, self.transpose(Z)), Jmat))

    def euclidean_to_unitary(self, U: Matrix, dU_euclidean: Matrix) -> Matrix:
        r"""Convert the Euclidean gradient to a Riemannian gradient on the
        tangent bundle of the unitary manifold.

        Implemented from:
            Y Yao, F Miatto, N Quesada - arXiv preprint arXiv:2209.06069, 2022.

        Args:
            U (Matrix): unitary matrix
            dU_euclidean (Matrix): Euclidean gradient tensor

        Returns:
            Matrix: unitary gradient tensor
        """
        Z = self.matmul(self.conj(self.transpose(U)), dU_euclidean)
        return 0.5 * (Z - self.conj(self.transpose(Z)))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of the class :class:`FockMeasurement`."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Sequence, Union

from mrmustard import math, settings
from mrmustard.math.parameter_set import ParameterSet
from mrmustard.math.parameters import Constant, Variable
from mrmustard.utils.typing import Tensor
from .state import State


class Measurement(ABC):
    """this is an abstract class holding the common methods and properties that any measurement should
    implement

    Args:
        outcome (optional, List[float] or Tensor): the result of the measurement
        modes (List[int]): the modes on which the measurement is acting on
    """

    def __init__(self, outcome: Tensor, modes: Sequence[int]) -> None:
        super().__init__()

        if modes is None:
            raise ValueError(f"Modes not defined for {self.__class__.__name__}.")
        self._modes = modes

        self._is_postselected = False if outcome is None else True
        """used to evaluate if the measurement outcome should be
        sampled or is already defined by the user (postselection)"""

        self._parameter_set = ParameterSet()

    def _add_parameter(self, parameter: Union[Constant, Variable]):
        r"""
        Adds a parameter to a transformation.

        Args:
            parameter: The parameter to add.
        """
        self.parameter_set.add_parameter(parameter)
        self.__dict__[parameter.name] = parameter

    @property
    def parameter_set(self):
        r"""
        The set of parameters for this transformation.
        """
        return self._parameter_set

    @property
    def modes(self):
        r"""returns the modes being measured"""
        return self._modes

    @property
    def num_modes(self):
        r"""returns the number of modes being measured"""
        return len(self.modes)

    @property
    def postselected(self):
        r"""returns whether the measurement is postselected, i.e, a outcome has been provided"""
        return self._is_postselected

    @property
    @abstractmethod
    def outcome(self):
        """Returns outcome of the measurement. If no measurement has been carried out returns `None`."""
        ...

    @abstractmethod
    def _measure_fock(self, other: State) -> Union[State, float]: ...

    @abstractmethod
    def _measure_gaussian(self, other: State) -> Union[State, float]: ...

    def primal(self, other: State) -> Union[State, float]:
        """performs the measurement procedure according to the representation of the incoming state"""
        if other.is_gaussian:
            return self._measure_gaussian(other)

        return self._measure_fock(other)

    def __lshift__(self, other) -> Union[State, float]:
        if isinstance(other, State):
            self.primal(other)

        raise TypeError(
            f"Cannot apply Measurement '{self.__qualname__}' to '{other.__qualname__}'."
        )

    def __getitem__(self, items) -> Measurement:
        """Assign modes via the getitem syntax: allows measurements to be used as
        ``output = meas[0,1](input)``, e.g. measuring modes 0 and 1.
        """
        if isinstance(items, int):
            modes = [items]
        elif isinstance(items, slice):
            modes = list(range(items.start, items.stop, items.step))
        elif isinstance(items, (Sequence, Iterable)):
            modes = list(items)
        else:
            raise ValueError(f"{items} is not a valid slice or list of modes.")
        self._modes = modes

        return self


class FockMeasurement(Measurement):
    """A Fock measurement projecting onto a Fock measurement pattern.

    It works by representing the state in the Fock basis and then applying a stochastic channel
    matrix ``P(meas|n)`` to the Fock probabilities (belief propagation).

    It outputs the measurement probabilities and the remaining post-measurement state (if any)
    in the Fock basis.
    """

    def __init__(self, outcome: Tensor, modes: Sequence[int], cutoffs: Sequence[int]) -> None:
        self._cutoffs = cutoffs or [settings.PNR_INTERNAL_CUTOFF] * len(modes)
        super().__init__(outcome, modes)

    @property
    def outcome(self):
        raise NotImplementedError

    def _measure_gaussian(self, other: State) -> Union[State, float]:
        return self._measure_fock(other)

    def _measure_fock(self, other: State) -> Union[State, float]:
        r"""
        Returns a tensor representing the post-measurement state in the unmeasured modes in the Fock basis.
        The first `N` indices of the returned tensor correspond to the Fock measurements of the `N` modes that
        the detector is measuring. The remaining indices correspond to the density matrix of the unmeasured modes.

        Args
            other (State): the quantum state
        Returns
            Tensor: a tensor representing the post-measurement state
        """
        cutoffs = []
        for mode in other.modes:
            if mode in self._modes:
                cutoffs.append(
                    max(settings.PNR_INTERNAL_CUTOFF, other.cutoffs[other.indices(mode)])
                )
            else:
                cutoffs.append(other.cutoffs[other.indices(mode)])
        if self.should_recompute_stochastic_channel() or math.any(
            [c > settings.PNR_INTERNAL_CUTOFF for c in other.cutoffs]
        ):
            self.recompute_stochastic_channel(cutoffs)
        dm = other.dm(cutoffs)
        for k, (mode, stoch) in enumerate(zip(self._modes, self._internal_stochastic_channel)):
            # move the mode indices to the end
            last = [mode - k, mode + other.num_modes - 2 * k]
            perm = [m for m in range(dm.ndim) if m not in last] + last
            dm = math.transpose(dm, perm)
            # compute sum_m P(meas|m)rho_mm
            dm = math.diag_part(dm)
            dm = math.tensordot(dm, stoch[: self._cutoffs[k], : dm.shape[-1]], [[-1], [1]])
        # put back the last len(self.modes) modes at the beginning
        output = math.transpose(
            dm,
            list(range(dm.ndim - len(self._modes), dm.ndim))
            + list(range(dm.ndim - len(self._modes))),
        )
        if len(output.shape) == len(self._modes):  # all modes are measured
            output = math.real(output)  # return probabilities
        return output

    #  pylint: disable=no-self-use
    def should_recompute_stochastic_channel(self) -> bool:  # override in subclasses
        """Returns `True` if the stochastic channel has to be recomputed.

        This method should be overriden by subclasses as needed.
        """
        return False

    def recompute_stochastic_channel(self, cutoffs: Sequence[int]) -> None:
        """Recomputes the stochastic channel.

        This method should be overriden by subclasses as needed.
        """
        raise NotImplementedError



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of optimization classes and functions
used within Mr Mustard.
"""

from itertools import chain, groupby
from typing import List, Callable, Sequence, Union, Mapping, Dict
from mrmustard import math
from mrmustard.math.parameters import Constant, Variable
from mrmustard.training.callbacks import Callback
from mrmustard.training.progress_bar import ProgressBar
from mrmustard.utils.logger import create_logger
from mrmustard.math.parameters import (
    update_euclidean,
    update_orthogonal,
    update_symplectic,
    update_unitary,
)
from mrmustard.lab import Circuit

__all__ = ["Optimizer"]


# pylint: disable=disallowed-name
class Optimizer:
    r"""An optimizer for any parametrized object: it can optimize euclidean, orthogonal and symplectic parameters.

    .. note::

        In the future it will also include a compiler, so that it will be possible to
        simplify the circuit/detector/gate/etc before the optimization and also
        compile other types of structures like error correcting codes and encoders/decoders.
    """

    def __init__(
        self,
        symplectic_lr: float = 0.1,
        unitary_lr: float = 0.1,
        orthogonal_lr: float = 0.1,
        euclidean_lr: float = 0.001,
    ):
        self.learning_rate = {
            update_euclidean: euclidean_lr,
            update_symplectic: symplectic_lr,
            update_unitary: unitary_lr,
            update_orthogonal: orthogonal_lr,
        }
        self.opt_history: List[float] = [0]
        self.callback_history: Dict[str, List] = {}
        self.log = create_logger(__name__)

    def minimize(
        self,
        cost_fn: Callable,
        by_optimizing: Sequence[Union[Constant, Variable, Circuit]],
        max_steps: int = 1000,
        callbacks: Union[Callable, Sequence[Callable], Mapping[str, Callable]] = None,
    ):
        r"""Minimizes the given cost function by optimizing circuits and/or detectors.

        Args:
            cost_fn (Callable): a function that will be executed in a differentiable context in
                order to compute gradients as needed
            by_optimizing (list of circuits and/or detectors and/or gates): a list of elements that
                contain the parameters to optimize
            max_steps (int): the minimization keeps going until the loss is stable or max_steps are
                reached (if ``max_steps=0`` it will only stop when the loss is stable)
            callbacks (:class:`Callback`, `Callable`, or List/Dict of them): callback functions that
                will be executed at each step of the optimization after backprop but before gradient
                gets applied. It takes as arguments the optimizer itself, training step (int), the
                cost value, the cost function, and the trainable parameters (values & grads) dict.
                The optional returned dict for each step is stored in self.callback_history which
                is a callback-name-keyed dict with each value a list of such callback result dicts.
                Learn more about how to use callbacks to have finer control of the optimization
                process in the :mod:`.callbacks` module.
        """
        callbacks = self._coerce_callbacks(callbacks)

        try:
            self._minimize(cost_fn, by_optimizing, max_steps, callbacks)
        except KeyboardInterrupt:  # graceful exit
            self.log.info("Optimizer execution halted due to keyboard interruption.")
            raise self.OptimizerInterruptedError() from None

    def _minimize(self, cost_fn, by_optimizing, max_steps, callbacks):
        # finding out which parameters are trainable from the ops
        trainable_params = self._get_trainable_params(by_optimizing)
        cost_fn_modified = False
        orig_cost_fn = cost_fn

        bar = ProgressBar(max_steps)
        with bar:
            while not self.should_stop(max_steps):
                cost, grads = self.compute_loss_and_gradients(cost_fn, trainable_params.values())

                trainables = {tag: (x, dx) for (tag, x), dx in zip(trainable_params.items(), grads)}

                if cost_fn_modified:
                    self.callback_history["orig_cost"].append(orig_cost_fn())

                new_cost_fn, new_grads = self._run_callbacks(
                    callbacks=callbacks,
                    cost_fn=cost_fn,
                    cost=cost,
                    trainables=trainables,
                )

                self.apply_gradients(trainable_params.values(), new_grads or grads)
                self.opt_history.append(cost)
                bar.step(math.asnumpy(cost))

                if callable(new_cost_fn):
                    cost_fn = new_cost_fn
                    if not cost_fn_modified:
                        cost_fn_modified = True
                        self.callback_history["orig_cost"] = self.opt_history.copy()

    def apply_gradients(self, trainable_params, grads):
        """Apply gradients to variables.

        This method group parameters by variable type (euclidean, symplectic, orthogonal) and
        applies the corresponding update method for each variable type. Update methods are
        registered on :mod:`parameter_update` module.
        """
        grouped_items = sorted(
            zip(grads, trainable_params),
            key=lambda x: hash(getattr(x[1], "update_fn", update_euclidean)),
        )
        grouped_items = {
            key: list(result)
            for key, result in groupby(
                grouped_items, key=lambda x: hash(getattr(x[1], "update_fn", update_euclidean))
            )
        }

        for grads_vars in grouped_items.values():
            update_fn = getattr(grads_vars[0][1], "update_fn", update_euclidean)
            params_lr = self.learning_rate[update_fn]
            # extract value (tensor) from the parameter object and group with grad
            grads_and_vars = [(grad, p.value) for grad, p in grads_vars]
            update_fn(grads_and_vars, params_lr)

    @staticmethod
    def _get_trainable_params(trainable_items, root_tag: str = "optimized"):
        """Traverses all instances of gates, states, detectors, or trainable items that belong to the backend
        and return a dict of trainables of the form `{tags: trainable_parameters}` where the `tags`
        are traversal paths of collecting all parent tags for reaching each parameter.
        """
        trainables = []
        for i, item in enumerate(trainable_items):
            owner_tag = f"{root_tag}[{i}]"
            if isinstance(item, Circuit):
                for j, op in enumerate(item.ops):
                    tag = f"{owner_tag}:{item.__class__.__qualname__}/_ops[{j}]"
                    tagged_vars = op.parameter_set.tagged_variables(tag)
                    trainables.append(tagged_vars.items())
            elif hasattr(item, "parameter_set"):
                tag = f"{owner_tag}:{item.__class__.__qualname__}"
                tagged_vars = item.parameter_set.tagged_variables(tag)
                trainables.append(tagged_vars.items())
            elif math.from_backend(item) and math.is_trainable(item):
                # the created parameter is wrapped into a list because the case above
                # returns a list, hence ensuring we have a list of lists
                tag = f"{owner_tag}:{math.__class__.__name__}/{getattr(item, 'name', item.__class__.__name__)}"
                trainables.append([(tag, Variable(item, name="from _backend"))])

        return dict(chain(*trainables))

    @staticmethod
    def _group_vars_and_grads_by_type(trainable_params, grads):
        """Groups `trainable_params` and `grads` by type into a dict of the form
        `{"euclidean": [...], "orthogonal": [...], "symplectic": [...]}, "unitary": [...]`."""
        sorted_grads_and_vars = sorted(
            zip(grads, trainable_params), key=lambda grads_vars: grads_vars[1].type
        )
        grouped = {
            key: list(result)
            for key, result in groupby(
                sorted_grads_and_vars, key=lambda grads_vars: grads_vars[1].type
            )
        }

        return grouped

    @staticmethod
    def compute_loss_and_gradients(cost_fn: Callable, parameters: List[Variable]):
        r"""Uses the backend to compute the loss and gradients of the parameters
        given a cost function.

        This functions is a wrapper around the backend optimizer to extract tensors
        from `parameters` and correctly compute the loss and gradients. Results of
        the calculation are associated back again with the given parameters.

        Args:
            cost_fn (Callable with no args): The cost function.
            parameters (List[Variable]): The variables to optimize.

        Returns:
            tuple(Tensor, List[Tensor]): The loss and the gradients.
        """
        param_tensors = [p.value for p in parameters]
        loss, grads = math.value_and_gradients(cost_fn, param_tensors)

        return loss, grads

    def should_stop(self, max_steps: int) -> bool:
        r"""Returns ``True`` if the optimization should stop (either because
        the loss is stable or because the maximum number of steps is reached)."""
        if max_steps != 0 and len(self.opt_history) > max_steps:
            return True
        if len(self.opt_history) > 20:  # if cost varies less than 10e-6 over 20 steps
            if (
                sum(abs(self.opt_history[-i - 1] - self.opt_history[-i]) for i in range(1, 20))
                < 1e-6
            ):
                self.log.info("Loss looks stable, stopping here.")
                return True
        return False

    @staticmethod
    def _coerce_callbacks(callbacks):
        r"""Coerce callbacks into dict and validate them."""
        if callbacks is None:
            callbacks = {}
        elif callable(callbacks):
            callbacks = {
                callbacks.tag if isinstance(callbacks, Callback) else callbacks.__name__: callbacks
            }
        elif isinstance(callbacks, Sequence):
            callbacks = {
                cb.tag if isinstance(cb, Callback) else cb.__name__: cb for cb in callbacks
            }
        elif not isinstance(callbacks, Mapping):
            raise TypeError(
                f"Argument `callbacks` expected to be a callable or a list/dict of callables, got {type(callbacks)}."
            )

        if any(not callable(cb) for cb in callbacks.values()):
            raise TypeError("Not all provided callbacks is callable.")

        return callbacks

    def _run_callbacks(self, callbacks, cost_fn, cost, trainables):
        """Iteratively calls all callbacks and applies the necessary updates."""
        new_cost_fn, new_grads = None, None

        for cb_tag, cb in callbacks.items():
            if cb_tag not in self.callback_history:
                self.callback_history[cb_tag] = []

            cb_result = cb(
                optimizer=self,
                cost_fn=cost_fn if new_cost_fn is None else new_cost_fn,
                cost=cost,
                trainables=trainables,
            )

            if not isinstance(cb_result, (Mapping, type(None))):
                raise TypeError(
                    f"The expected return type of callback functions is dict, got {type(cb_result)}."
                )

            new_cost_fn = cb_result.pop("cost_fn", None)

            if "grads" in cb_result:
                new_grads = cb_result["grads"]
                trainables = {
                    tag: (x, dx) for (tag, (x, _)), dx in zip(trainables.items(), new_grads)
                }

            if cb_result is not None and cb_result:
                self.callback_history[cb_tag].append(cb_result)

        return new_cost_fn, new_grads

    class OptimizerInterruptedError(Exception):
        """A helper class to quietly stop execution without printing a traceback."""

        def _render_traceback_(self):
            pass



"""
This module calculates the diagonal of the Fock representation (i.e. the PNR detection probabilities of all modes)
by applying the recursion relation in a selective manner.
"""

import numpy as np
import numba
from numba import njit, int64
from numba.cpython.unsafe.tuple import tuple_setitem
from mrmustard.math.lattice.strategies.compactFock.helperFunctions import (
    SQRT,
    repeat_twice,
    construct_dict_params,
)


@njit
def use_offDiag_pivot(
    A, B, M, cutoffs, params, d, arr0, arr2, arr1010, arr1001, arr1
):  # pragma: no cover
    """
    Apply recurrence relation for pivot of type [a+1,a,b,b,c,c,...] / [a,a,b+1,b,c,c,...] / [a,a,b,b,c+1,c,...]
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        d (int): mode index in which the considered Fock amplitude is off diagonal
            e.g. [a,a,b+1,b,c,c,...] --> b is off diagonal --> d=1
        arr0, arr2, arr1010, arr1001, arr1 (array, array, array, array, array): submatrices of the fock representation
    Returns:
        (array, array, array, array, array): updated versions of arr0, arr2, arr1010, arr1001, arr1
    """
    pivot = repeat_twice(params)
    pivot[2 * d] += 1
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    if B.ndim == 1:
        G_in = np.zeros(2 * M, dtype=np.complex128)
    elif B.ndim == 2:
        G_in = np.zeros((2 * M, B.shape[1]), dtype=np.complex128)

    ########## READ ##########
    GB = arr1[(2 * d,) + params] * B

    # Array0
    G_in[2 * d] = arr0[params]

    # read from Array2
    if params[d] > 0:
        params_adapted = tuple_setitem(params, d, params[d] - 1)
        G_in[2 * d + 1] = arr2[(d,) + params_adapted]

    # read from Array11
    for i in range(d + 1, M):  # i>d
        if params[i] > 0:
            params_adapted = tuple_setitem(params, i, params[i] - 1)
            G_in[2 * i] = arr1001[(d, i - d - 1) + params_adapted]
            G_in[2 * i + 1] = arr1010[(d, i - d - 1) + params_adapted]

    ########## WRITE ##########
    if B.ndim == 1:
        G_in = np.multiply(K_l, G_in)
    elif B.ndim == 2:
        G_in = np.multiply(np.expand_dims(K_l, 1), G_in)

    # Array0
    params_adapted = tuple_setitem(params, d, params[d] + 1)
    arr0[params_adapted] = (GB[2 * d + 1] + A[2 * d + 1] @ G_in) / K_i[2 * d + 1]

    # Array2
    if params[d] + 2 < cutoffs[d]:
        arr2[(d,) + params] = (GB[2 * d] + A[2 * d] @ G_in) / K_i[2 * d]

    # Array11
    for i in range(d + 1, M):
        if params[i] + 1 < cutoffs[i]:
            arr1010[(d, i - d - 1) + params] = (GB[2 * i] + A[2 * i] @ G_in) / K_i[2 * i]
            arr1001[(d, i - d - 1) + params] = (GB[2 * i + 1] + A[2 * i + 1] @ G_in) / K_i[
                2 * i + 1
            ]

    return arr0, arr2, arr1010, arr1001


@njit
def use_diag_pivot(A, B, M, cutoffs, params, arr0, arr1):  # pragma: no cover
    """
    Apply recurrence relation for pivot of type [a,a,b,b,c,c...]
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        arr0, arr1 (array, array): submatrices of the fock representation
    Returns:
        (array, array): updated versions of arr0, arr1
    """
    pivot = repeat_twice(params)
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    if B.ndim == 1:
        G_in = np.zeros(2 * M, dtype=np.complex128)
    elif B.ndim == 2:
        G_in = np.zeros((2 * M, B.shape[1]), dtype=np.complex128)

    ########## READ ##########
    GB = arr0[params] * B

    # Array1
    for i in range(2 * M):
        if params[i // 2] > 0:
            params_adapted = tuple_setitem(params, i // 2, params[i // 2] - 1)
            G_in[i] = arr1[
                (i + 1 - 2 * (i % 2),) + params_adapted
            ]  # [i+1-2*(i%2) for i in range(6)] = [1,0,3,2,5,4]

    ########## WRITE ##########
    if B.ndim == 1:
        G_in = np.multiply(K_l, G_in)
    elif B.ndim == 2:
        G_in = np.multiply(np.expand_dims(K_l, 1), G_in)

    # Array1
    for i in range(2 * M):
        if params[i // 2] + 1 < cutoffs[i // 2]:
            # this prevents a few elements from being written that will never be read
            if i != 1 or params[0] + 2 < cutoffs[0]:
                arr1[(i,) + params] = (GB[i] + A[i] @ G_in) / K_i[i]

    return arr1


@njit
def fock_representation_diagonal_amps_NUMBA(
    A, B, M, cutoffs, arr0, arr2, arr1010, arr1001, arr1, tuple_type, list_type
):  # pragma: no cover
    """
    Returns the PNR probabilities of a mixed state according to algorithm 1 of:
    https://doi.org/10.22331/q-2023-08-29-1097
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        arr0 (array): submatrix of the fock representation that contains Fock amplitudes of the type [a,a,b,b,c,c...]
            (!) should already contain G0 at position (0,)*M
        arr2 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+2,a,b,b,c,c...] / [a,a,b+2,b,c,c...] / ...
        arr1010 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b+1,b,c,c,...] / [a+1,a,b,b,c+1,c,...] / [a,a,b+1,b,c+1,c,...] / ...
        arr1001 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b+1,c,c,...] / [a+1,a,b,b,c,c+1,...] / [a,a,b+1,b,c,c+1,...] / ...
        arr1 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b,c,c...] / [a,a+1,b,b,c,c...] / [a,a,b+1,b,c,c...] / ...
        tuple_type, list_type (numba types): numba types that need to be defined outside of numba compiled functions
    Returns:
        array: the fock representation
    """
    dict_params = construct_dict_params(cutoffs, tuple_type, list_type)
    for sum_params in range(sum(cutoffs)):
        for params in dict_params[sum_params]:
            # diagonal pivots: aa,bb,cc,dd,...
            if (cutoffs[0] == 1) or (params[0] < cutoffs[0] - 1):
                arr1 = use_diag_pivot(A, B, M, cutoffs, params, arr0, arr1)
            # off-diagonal pivots: d=0: (a+1)a,bb,cc,dd,... | d=1: 00,(b+1)b,cc,dd | 00,00,(c+1)c,dd | ...
            for d in range(M):
                if np.all(np.array(params)[:d] == 0) and (params[d] < cutoffs[d] - 1):
                    arr0, arr2, arr1010, arr1001 = use_offDiag_pivot(
                        A, B, M, cutoffs, params, d, arr0, arr2, arr1010, arr1001, arr1
                    )
    return arr0, arr2, arr1010, arr1001, arr1


def fock_representation_diagonal_amps(A, B, G0, M, cutoffs):
    """
    First initialise the submatrices of G (of which the shape depends on cutoff and M)
    and some other constants
    (These initialisations currently cannot be done using Numba.)
    Then calculate the fock representation.
    """

    cutoffs = tuple(cutoffs)
    tuple_type = numba.types.UniTuple(int64, M)
    list_type = numba.types.ListType(tuple_type)

    if B.ndim == 1:
        arr0 = np.zeros(cutoffs, dtype=np.complex128)
        arr2 = np.zeros((M,) + cutoffs, dtype=np.complex128)
        arr1 = np.zeros((2 * M,) + cutoffs, dtype=np.complex128)
        if M == 1:
            arr1010 = np.zeros((1, 1, 1), dtype=np.complex128)
            arr1001 = np.zeros((1, 1, 1), dtype=np.complex128)
        else:
            arr1010 = np.zeros((M, M - 1) + cutoffs, dtype=np.complex128)
            arr1001 = np.zeros((M, M - 1) + cutoffs, dtype=np.complex128)

    elif B.ndim == 2:
        batch_length = B.shape[1]
        arr0 = np.zeros(cutoffs + (batch_length,), dtype=np.complex128)
        arr2 = np.zeros((M,) + cutoffs + (batch_length,), dtype=np.complex128)
        arr1 = np.zeros((2 * M,) + cutoffs + (batch_length,), dtype=np.complex128)
        if M == 1:
            arr1010 = np.zeros((1, 1, 1) + (batch_length,), dtype=np.complex128)
            arr1001 = np.zeros((1, 1, 1) + (batch_length,), dtype=np.complex128)
        else:
            arr1010 = np.zeros((M, M - 1) + cutoffs + (batch_length,), dtype=np.complex128)
            arr1001 = np.zeros((M, M - 1) + cutoffs + (batch_length,), dtype=np.complex128)

    arr0[(0,) * M] = G0
    return fock_representation_diagonal_amps_NUMBA(
        A, B, M, cutoffs, arr0, arr2, arr1010, arr1001, arr1, tuple_type, list_type
    )
